#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2020"
__version__		= "0.2.4"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	Motifs
#
#
#	Copyright (c) Sander Granneman 2020
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import math
import time
from random import randint
from collections import defaultdict
from pyCRAC.Classes.NGSFormatWriters import *
from pyCRAC.Parsers import ParseAlignments
from pyCRAC.Methods import sortbyvalue,orientation,numpy_overlap,getfilename,check_list,reverse_complement

def randomizeSequences(orientation,sequence,gene_sequence):
	"""Returns a random sequence from a particular gene. If the read sequence is in the
	anti-sense orientation, the reverse_complement of a random sequence will be returned"""
	rand_start = 0
	rand_end   = 0
	limit	   = len(gene_sequence)-len(sequence)
	if len(gene_sequence)-len(sequence) + 1 > 2:					# The difference in length between the gene sequence and the read sequence has to be at least two nucleotides
		rand_start = randint(1,limit)								# Otherwise, randint will only be able to generate '1' as random number
		rand_end   = rand_start + len(sequence)
	else:															# If the gene sequence is smaller than the actual read sequence (think tRNAs, small snoRNAs, etc) then
		rand_start = 1												# randint gets a negative number. This obviously won't work so in these cases the gene sequences are simply returned
		rand_end   = len(gene_sequence)								# It is important that tRNAs are not (always) included in the motif searches since these are very similar and highest scoring motifs would
	if orientation == "sense":										# therefore almost always come from tRNAs.
		rand_seq   = gene_sequence[rand_start:rand_end]
		return rand_seq
	elif orientation == "anti-sense":
		rand_seq   = reverse_complement(gene_sequence)[rand_start:rand_end]
		return rand_seq
	else:
		sys.stderr.write("could not return a random sequence from your input\nPlease check your input\n")

class FindMotifs():
	"""This class extracts motifs from intervals, calculates the Z-score for each motif found and prints a file where in genes
	these motifs were found. A file containing the hits for each motifs in the reads and a file with hits for randomly
	generated motifs will also be printed. Requires a GTF2 object to find overlap between intervals and genomic features """
	def __init__(self):
		self.__genomic_seq		= str()
		self.__gene_rand_seq	= list()
		self.__rand_count		= defaultdict(int)
		self.__motif_dict		= defaultdict(set)
		self.__motif_count		= defaultdict(int)
		self.__rand_motif_count = defaultdict(int)
		self.__k_length_range	= list()
		self.__insert			= str()
		self.__insert2			= str()
		self.__gtf				= None
		self.annotation			= str()
		self.total_intervals	= 0
		self.z_scores			= defaultdict(int)
		self.file_name			= str()
		self.file_list			= list()

	def setKmerLength(self,k_min=4,k_max=8):
		"""To change the default k-mer length settings (defaults are between 4 and 8 nucleotides)"""
		if k_max > k_min:
			self.__k_length_range = range(k_min,(k_max+1))
		elif k_max < k_min:										# in case people flip the two accidentally
			self.__k_length_range = range(k_max,(k_min+1))
		else:
			self.__k_length_range = [k_min]						# in k_min and k_max are the same

	def findMotifs(self,gtf,datafile,min_overlap=1):
		"""Extracts the motifs from intervals."""
		#-----------------------------------------------------------------------------------------------------------------------------------

		search_results	  = list()
		motif_list		  = list()
		list_of_tuples	  = defaultdict(list)
		self.__gtf		  = gtf

		#------------ extracting motifs from each read coordinate. Anti-sense hits are considered but unmapped reads are not ---------------

		data = ParseAlignments.ParseCountersOutput(datafile)
		self.file_name = getfilename(datafile)
		while data.readLineByLine(collectmutations=False):
			data.duplicatesremoved = True
			self.total_intervals += 1
			read_sequence = gtf.sequence(data.chromosome,data.strand,data.read_start,data.read_end)
			motif_start = int()
			motif_end	= int()
			readorientation = str()
			overlapping_genes = list()
			for gene in data.genes:
				if gene == "no_matches":		# if the interval has not been mapped to a genomic feature, ignore it
					continue
				try:
					readorientation = orientation(data.strand,gtf.strand(gene))
				except LookupError:
					continue
				if readorientation == "sense":																	# If the intervals are sense to the gene or GTF feature
					if gtf.strand(gene) == "+":
						for k_length in self.__k_length_range:
							length_seq = len(read_sequence) - k_length
							for i in range (0,length_seq+1):
								motif = read_sequence[i:(i+k_length)]
								if check_list(motif_list,motif):												# checks if the motif sequence has already been seen in the read or cluster.
									motif_start = data.read_start + i
									self.__motif_count[motif] += 1
									self.__motif_dict[motif].add((data.chromosome,data.strand,motif_start))
							motif_list = list()

					elif gtf.strand(gene) == "-":
						for k_length in self.__k_length_range:
							length_seq = len(read_sequence) - k_length
							for i in range (0,length_seq+1):
								motif = read_sequence[i:(i+k_length)]
								if check_list(motif_list,motif):
									motif_start = data.read_end - k_length - i
									self.__motif_count[motif] += 1
									self.__motif_dict[motif].add((data.chromosome,data.strand,motif_start))
							motif_list = list()

				elif readorientation == "anti-sense":															# If the intervals are anti-sense to the gene or GTF feature
					if gtf.strand(gene) == "-":
						for k_length in self.__k_length_range:
							length_seq = len(read_sequence) - k_length
							for i in range (0,length_seq+1):
								motif = read_sequence[i:(i+k_length)]
								if check_list(motif_list,motif):
									motif_start = data.read_start + i
									self.__motif_count[motif] += 1
									self.__motif_dict[motif].add((data.chromosome,data.strand,motif_start))
							motif_list = list()


					elif gtf.strand(gene) == "+":
						for k_length in self.__k_length_range:
							length_seq = len(read_sequence) - k_length
							for i in range (0,length_seq+1):
								motif = read_sequence[i:(i+k_length)]
								if check_list(motif_list,motif):
									motif_start = data.read_end - k_length - i
									self.__motif_count[motif] += 1
									self.__motif_dict[motif].add((data.chromosome,data.strand,motif_start))
							motif_list = list()


				random_sequence = randomizeSequences(readorientation,read_sequence,gtf.genomicSequence(gene,ranges=len(read_sequence)-min_overlap))		        # some of the reads may overlap with the 5' and 3' UTR, therefore,
				for k_length in self.__k_length_range:																											# for making random sequences the program includes a stretch
					length_seq = len(random_sequence) - k_length																								# of the 5' and 3' UTRs with the length of the read sequence length!
					for i in range (0,length_seq+1):
						rand_motif = random_sequence[i:(i+k_length)]
						if check_list(motif_list,rand_motif):
							self.__rand_count[rand_motif] += 1
					motif_list = list()

		return True

	def __createFileHandles(self,file_type,file_extension="txt"):
		""" does what it says """
		annotation = "all"
		if self.annotation:
			annotation = self.annotation
		filename = "%s_%s_%s.%s" % (self.file_name,annotation,file_type,file_extension)
		self.file_list.append(filename)
		return filename

	def printRandKmers(self,output_file=None,max_motifs=1000):
		"""Prints a list of hits for randomly generated k-mers"""
		assert self.__motif_count, "\n\nYou need to run the extractMotifs method first before you can print the k-mers\n"
		file_name = None
		file_type = "random_k-mers_count"
		if output_file:
			self.file_name = output_file
		file_name = self.__createFileHandles(file_type)
		file_out = open(file_name, "w")
		file_out.write("# %s\n# %s\n" % (' '.join(sys.argv),time.ctime()))
		file_out.write("# k-mer lengths:\n#\tmin: %s\n#\tmax: %s\n" % (self.__k_length_range[0],self.__k_length_range[-1]))
		file_out.write("# %s intervals\n\n# k-mer\tnumber_of_occurrences\n" % (self.total_intervals))
		for (motifs,value) in sortbyvalue(self.__rand_count)[0:max_motifs]:
			file_out.write("%s\t%s\n" % (motifs,self.__rand_count[motifs]))
		file_out.close()

	def printExperimentalKmers(self,output_file=None,max_motifs=1000):
		"""Prints a k-mer hit list extracted from the data"""
		assert self.__motif_count, "\n\nYou need to run the extractMotifs method first before you can print the k-mers\n"
		file_name = None
		file_type = "data_k-mers_count"
		if output_file:
			self.file_name = output_file
		file_name = self.__createFileHandles(file_type)
		file_out = open(file_name, "w")
		file_out.write("# %s\n# %s\n" %(' '.join(sys.argv),time.ctime()))
		file_out.write("# k-mer lengths:\n#\tmin: %s\n#\tmax: %s\n" % (self.__k_length_range[0],self.__k_length_range[-1]))
		file_out.write("# %s intervals\n\n# k-mer\tnumber_of_occurrences\n" % self.total_intervals)
		for (motifs,value) in sortbyvalue(self.__motif_count)[0:max_motifs]:
			file_out.write("%s\t%s\n" % (motifs,self.__motif_count[motifs]))
		file_out.close()

	def calcMotifZscores(self):
		"""Calculates the Z-scores for each k-mer sequence"""
		assert self.__motif_count, "\n\nYou need to run the extractMotifs method first before you can calculate Z-scores\n"
		motif_list = list()
		a = set(self.__motif_count.keys())
		b = set(self.__rand_count.keys())
		sample_sum = sum(self.__motif_count.values())
		control_sum = sum(self.__rand_count.values())
		motif_list	 = a | b
		for motifs in motif_list:
			p1 = float(self.__motif_count[motifs]) / float(sample_sum)
			p2 = float(self.__rand_count[motifs]) / float(control_sum)
			try:
				if p1 or p2:
					Z_score = math.sqrt(sample_sum) * (p1-p2) / math.sqrt(p1*(1-p1) + p2*(1-p2))
					self.z_scores[motifs] = Z_score
			except ValueError:
				sys.stderr.write("\nValueError! Can not calculate a Z-score for the %s motif. The number of motifs (%s) is higher than the total number of sequences (%s)\n" % (motifs,self.__motif_count[motifs],sample_sum))
				pass
		return True

	def printMotifZscores(self,output_file=None,max_motifs=1000):
		"""Prints the Z-scores for the top k-mers"""
		assert self.__motif_count, "\n\nYou need to run the extractMotifs method first before you can print the Z-scores\n"
		file_type = "k-mer_Z_scores"
		file_name = None
		if output_file:
			self.file_name = output_file
		file_name = self.__createFileHandles(file_type)
		file_out = open(file_name,"w")
		file_out.write("# %s\n# %s\n" % (' '.join(sys.argv),time.ctime()))
		file_out.write("# k-mer lengths:\n#\tmin: %s\n#\tmax: %s\n" % (self.__k_length_range[0],self.__k_length_range[-1]))
		file_out.write("# %s intervals\n\n# k-mer\tZ-score\n" % (self.total_intervals))
		for (motifs,score) in sortbyvalue(self.z_scores)[0:max_motifs]:
			file_out.write("%s\t%0.2f\n" % (motifs,score))
		file_out.close()

	def printMotifGTFfile(self,output_file=None,max_motifs=1000):
		""" Prints a list of k-mers found in reads mapped to genes in GTF format"""
		assert self.__motif_count, "\n\nYou need to run the extractMotifs method first before you can print the GTF output file\n"
		list_of_tuples = defaultdict(list)
		if output_file:
			self.file_name = output_file
		if not self.z_scores:
			self.calcMotifZscores()
		topkmers  = sortbyvalue(self.z_scores)[0:max_motifs]
		file_name = None
		file_type = "top_k-mers_in_features"
		file_extension = "gtf"
		if output_file: file_name = output_file
		file_name = self.__createFileHandles(file_type,file_extension)
		file_out = NGSFileWriter(file_name)
		file_out.write("##gff-version 2\n")
		file_out.write("# %s\n# %s\n" %(' '.join(sys.argv),time.ctime()))
		file_out.write("# k-mer lengths:\n#\tmin: %s\n#\tmax: %s\n" % (self.__k_length_range[0],self.__k_length_range[-1]))
		file_out.write("# total number of unique k-mers reported:\t%s\n" % (max_motifs))
		file_out.write("# total number of mapped intervals:\t%s\n" % self.total_intervals)
		file_out.write("# chromosome\tsource\tsequence\tstart\tend\tZ-score\tstrand\t.\tattributes\n")
		for motif,value in topkmers:
			for interval in sorted(self.__motif_dict[motif]):
				chromosome,strand,motif_start = interval
				motif_length = len(motif)
				motif_end	 = motif_start+motif_length
				if chromosome not in list_of_tuples:
					try:
						list_of_tuples[chromosome] = self.__gtf.chromosomeGeneCoordIterator(chromosome,numpy=True)
					except AssertionError:
						continue
				genes = numpy_overlap(list_of_tuples[chromosome],motif_start,motif_end-1,overlap=1)
				if genes:
					gene_name = ",".join(genes)
					gene_id = ",".join([self.__gtf.gene2orf(gene) for gene in genes])
				else:
					gene_name = "no_matches"
					gene_id = "no_matches"
				file_out.writeGTF(chromosome,"motif",motif,motif_start,motif_end,value,strand=strand,gene_name=gene_name,gene_id=gene_id)
		file_out.close()



#!/usr/bin/python

__author__ = "Sander Granneman"
__copyright__ = "Copyright 2018"
__version__ = "0.0.5"
__credits__ = ["Sander Granneman"]
__maintainer__ = "Sander Granneman"
__email__ = "sgrannem@staffmail.ed.ac.uk"
__status__ = "Production"

##################################################################################
#
#	tab2dict.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import time
from collections import defaultdict

class Tab2Dict():
	"""this class can be used to store the information in a fasta sequence file into a dictionary.
	The dictionary keys are the gene or chromosome names, whereas the "values" is a single string of sequence.
	When parsing a large genomic reference sequence file, you can choose to put only a handful of chromosomes into memory
	by using the id argument followed by a list of the chromosome names"""

	def __init__(self,tab_file,ids=[]):
		self.__sequence = defaultdict(str)
		id =  str()
		if hasattr(tab_file,'read'):
			tab = tab_file
		else:
			tab = open(tab_file,"r")
		for line in tab:
			Fld = line.strip("\n").split('\t')
			id	= Fld[0].split(" ")[0]
			if not ids:
				self.__sequence[id] = Fld[1]
			else:
				if id in ids: self.__sequence[id] = Fld[1]

	def __getitem__(self,chromosome):
		return self.__sequence[chromosome]

	def __iter__(self):
		return iter(self.__sequence)

	def __setitem__(self,chromosome,sequence):
		self.__sequence[chromosome] = sequence

	def __call__(self):
		for key in self.__sequence:
			sys.stdout.write("%s\t%s\n" % (key,self.__sequence[key]))



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.6"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "beta"

##################################################################################
#
#	pyCheckGTFfile.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

##### This tool looks for duplicate gene names in GTF files and replaces these with the gene_id or transcript_name.

from pyCRAC.Classes.NGSFormatReaders import *
from pyCRAC.Methods import getfilename
from collections import defaultdict
from optparse import *

gtfdict = defaultdict(set)

def processGTFFile(infile,outfile,status):
	#store the gene_id-gene_name combinations in a dictionary and keeps track how many gene_ids there are per gene_name
	count = 0
	allowedfeatures = ["exon","CDS","5UTR","3UTR","start_codon","stop_codon"]
	gtf = NGSFileReader(infile)
	rejected = open("%s_incompatible.gtf" % getfilename(infile),"w")
	while gtf.readGTFLine():
		if gtf.feature in allowedfeatures:
			if not gtf.gene_name:
				gtf.gene_name = gtf.gene_id
			elif not gtf.gene_id:
				rejected.write(gtf.line)
				continue
			gtfdict[gtf.gene_name].add((gtf.chromosome,gtf.gene_id,gtf.gene_name))
	gtf = NGSFileReader(infile)
	outfile.write("##gff-version 2\n")
	while gtf.readGTFLine():
		if gtf.feature in allowedfeatures:
			if not gtf.gene_name:
				gtf.gene_name = gtf.gene_id
			if len(gtfdict[gtf.gene_name]) > 1:	   # if there is more than one gene_id for a gene_name, make sure the gene_name is changed
				if not gtf.gene_name:
					gtf.gene_name = gtf.gene_id					 
				if (gtf.chromosome,gtf.gene_id,gtf.gene_name) in gtfdict[gtf.gene_name]:
					position = sorted(gtfdict[gtf.gene_name]).index((gtf.chromosome,gtf.gene_id,gtf.gene_name))+1
					if gtf.gene_id == gtf.gene_name:
						gtf.line = gtf.line.replace("gene_id \"%s\"" % gtf.gene_id,"gene_id \"%s#%s\"" % (gtf.gene_id,position))
					gtf.line = gtf.line.replace("gene_name \"%s\"" % gtf.gene_name,"gene_name \"%s#%s\"" % (gtf.gene_name,position))
			outfile.write("%s" % gtf.line)
		else:
			rejected.write(gtf.line)
	outfile.close()
	rejected.close()

def main():
	parser = OptionParser(usage="usage: %prog [options] --gtf filename -o outputfile.gtf", version="%s" % __version__)
	parser.add_option("--gtf",dest="gtf_file",metavar="gtf input file",help="type the path to the gtf file that you want to use.",default=None)
	parser.add_option("-o","--output",dest="output_file",metavar="FILE",help="Optional. Specify the name of the output file. Default is standard output. Make sure it has the .gtf extension!",default=None)
	parser.add_option("-v","--verbose",dest="verbose",action="store_true",help="To print status messages to a log file",default=False)
	(options, args) = parser.parse_args()
	if len(sys.argv) < 1:
		parser.error("usage: %prog [options] --gtf filein -o fileout. Use -h or --help for options")
	if not options.gtf_file:
		parser.error("\nYou forgot to input the path to your GTF input file\n")
	if not options.output_file:
		parser.error("\nYou forgot to include a name for your output file\n")
	gtf = options.gtf_file
	outfile = open(options.output_file,"w")
	status = sys.stderr	 
	if options.verbose:
		status = open("pyCheckGTFfile_log.txt","w")
		
	processGTFFile(gtf,outfile,status)
		
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.1.0"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyGTF2sgr.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import numpy as np
from collections import defaultdict
from optparse import *
from pyCRAC.Parsers.ParseAlignments import ParseCountersOutput
from pyCRAC.Methods import getfilename,processChromFile
from pyCRAC.Classes.NGSFormatWriters import *

np.seterr(all="ignore")

def gtf2sgr(gtffile,chromosomedata,out_files=[],type="reads",permillion=False,score=False,log=sys.stdout,zeros=False,min_cov=0):
	"""Produces an sgr output file for hits, substitutions and deletions. It requires, besides data a file containing 
	chromosome names and chromosome lengths. Expects names for two output files in the out_files variable. The first file is assumed to be the output file
	for the plus strand. """
	typechoices = ["reads","substitutions","deletions","startpositions","endpositions"]
	assert type in typechoices, "The option %s is not supported\n" % type
	strands = ["+","-"]
	out = defaultdict()
	if not out_files: 
		out["+"] = NGSFileWriter()
		out["-"] = NGSFileWriter()
	else:
		out["+"] = NGSFileWriter(out_files[0])
		out["-"] = NGSFileWriter(out_files[1])

	data = ParseCountersOutput(gtffile)
	data.duplicatesremoved = True
	if score:
		data.duplicatesremoved = False
	current_chromosome = str()
	chromdict = defaultdict(list)
	
	### processing the chromosome data
	chromlengthfile = open(chromosomedata,"r").readlines()
	chromdata = processChromFile(chromlengthfile)
	###
	
	normvalue = int()

	while True:
		lines = data.readLineByLine()
		if current_chromosome and data.chromosome != current_chromosome or not lines:
			if permillion and not normvalue:
				if not data.mapped_reads:
					sys.stderr.write("No information on the number of mapped reads could be found in the gtf file. Cannot normalize the data\n")
					normvalue = 1.0
				else:
					normvalue = float(data.mapped_reads)/1000000.0 # to normalize the data to per million reads
			else:
				normvalue = 1
			for strand in strands:
				if min_cov:
					chromdict[strand][chromdict[strand] < min_cov] = 0.0
				if zeros:
					for position,hits in enumerate(chromdict[strand]):
						hits = float(hits)/normvalue
						out[strand].writeSgr(current_chromosome,position,hits)
				else:
					for position in np.where(chromdict[strand] > 0)[0]:
						hits = float(chromdict[strand][position])/normvalue
						out[strand].writeSgr(current_chromosome,position,hits)
			if not lines:
				break
			current_chromosome = str()
			
		if not current_chromosome:
			chromosome_length = chromdata[data.chromosome]
			for strand in strands:
				chromdict[strand] = np.zeros(chromosome_length)
			current_chromosome = data.chromosome
			log.write("Processing chromosome %s...\n" % current_chromosome)

		if type == "reads": 
			chromdict[data.strand][data.read_start:data.read_end] += data.number_of_reads
		elif type == "startpositions":
			if data.strand == "+":
				chromdict[data.strand][data.read_start] += data.number_of_reads
			else:
				chromdict[data.strand][data.read_end-1] += data.number_of_reads
		elif type == "endpositions":
			if data.strand == "-":
				chromdict[data.strand][data.read_start] += data.number_of_reads
			else:
				chromdict[data.strand][data.read_end-1] += data.number_of_reads
		elif type == "substitutions":
			if data.substitutions: chromdict[data.strand][data.substitutions] += data.number_of_reads
		elif type == "deletions":
			if data.deletions: chromdict[data.strand][data.deletions] += data.number_of_reads
			
def gtf2dropoffrates(gtffile,chromosomedata,out_files=[],score=False,log=sys.stdout,zeros=False,min_cov=0):
	"""Produces an sgr output file for hits, substitutions and deletions. It requires, besides data a file containing 
	chromosome names and chromosome lengths. Expects names for two output files in the out_files variable. The first file is assumed to be the output file
	for the plus strand. """
	strands = ["+","-"]
	out = defaultdict()
	if not out_files: 
		out["+"] = NGSFileWriter()
		out["-"] = NGSFileWriter()
	else:
		out["+"] = NGSFileWriter(out_files[0])
		out["-"] = NGSFileWriter(out_files[1])

	data = ParseCountersOutput(gtffile)
	data.duplicatesremoved = True
	if score:
		data.duplicatesremoved = False
	current_chromosome = str()
	
	### processing the chromosome data
	chromlengthfile = open(chromosomedata,"r").readlines()
	chromdata = processChromFile(chromlengthfile)
	###
	
	chromdict = defaultdict(lambda: defaultdict(list))
	
	while True:
		lines = data.readLineByLine()
		if current_chromosome and data.chromosome != current_chromosome or not lines:
			for strand in strands:
				if min_cov:
					chromdict[strand]["coverage"][chromdict[strand]["coverage"] < min_cov] = 0.0
					chromdict[strand]["fiveprimeend"][chromdict[strand]["fiveprimeend"] < min_cov] = 0.0
				dropoffs = np.nan_to_num(chromdict[strand]["fiveprimeend"]/chromdict[strand]["coverage"])
				if zeros:
					for position,hits in enumerate(dropoffs):
						out[strand].writeSgr(current_chromosome,position,hits)
				else:
					for position in np.where(dropoffs > 0)[0]:
						out[strand].writeSgr(current_chromosome,position,dropoffs[position])
			if not lines:
				break
			current_chromosome = str()
		if not current_chromosome:
			chromosome_length = chromdata[data.chromosome] + 1
			for strand in strands:
				chromdict[strand]["coverage"] = np.zeros(chromosome_length,dtype=float)
				chromdict[strand]["fiveprimeend"] = np.zeros(chromosome_length,dtype=float)
			current_chromosome = data.chromosome
			log.write("Processing chromosome %s...\n" % current_chromosome)

		chromdict[data.strand]["coverage"][data.read_start:data.read_end] += data.number_of_reads
		if data.strand == "+":
			chromdict[data.strand]["fiveprimeend"][data.read_start] += data.number_of_reads
		else:
			chromdict[data.strand]["fiveprimeend"][data.read_end-1] += data.number_of_reads
			
def gtf2endpositionrates(gtffile,chromosomedata,out_files=[],score=False,log=sys.stdout,zeros=False,min_cov=0):
	"""Produces an sgr output file for hits, substitutions and deletions. It requires, besides data a file containing 
	chromosome names and chromosome lengths. Expects names for two output files in the out_files variable. The first file is assumed to be the output file
	for the plus strand. """
	strands = ["+","-"]
	out = defaultdict()
	if not out_files: 
		out["+"] = NGSFileWriter()
		out["-"] = NGSFileWriter()
	else:
		out["+"] = NGSFileWriter(out_files[0])
		out["-"] = NGSFileWriter(out_files[1])

	data = ParseCountersOutput(gtffile)
	data.duplicatesremoved = True
	if score:
		data.duplicatesremoved = False
	current_chromosome = str()
	
	### processing the chromosome data
	chromlengthfile = open(chromosomedata,"r").readlines()
	chromdata = processChromFile(chromlengthfile)
	###
	
	chromdict = defaultdict(lambda: defaultdict(list))
	
	while True:
		lines = data.readLineByLine()
		if current_chromosome and data.chromosome != current_chromosome or not lines:
			for strand in strands:
				if min_cov:
					chromdict[strand]["coverage"][chromdict[strand]["coverage"] < min_cov] = 0.0
					chromdict[strand]["threeprimeend"][chromdict[strand]["threeprimeend"] < min_cov] = 0.0
				dropoffs = np.nan_to_num(chromdict[strand]["threeprimeend"]/chromdict[strand]["coverage"])
				if zeros:
					for position,hits in enumerate(dropoffs):
						out[strand].writeSgr(current_chromosome,position,hits)
				else:
					for position in np.where(dropoffs > 0)[0]:
						out[strand].writeSgr(current_chromosome,position,dropoffs[position])
			if not lines:
				break
			current_chromosome = str()
		if not current_chromosome:
			chromosome_length = chromdata[data.chromosome] + 1
			for strand in strands:
				chromdict[strand]["coverage"] = np.zeros(chromosome_length,dtype=float)
				chromdict[strand]["threeprimeend"] = np.zeros(chromosome_length,dtype=float)
			current_chromosome = data.chromosome
			log.write("Processing chromosome %s...\n" % current_chromosome)

		chromdict[data.strand]["coverage"][data.read_start:data.read_end] += data.number_of_reads
		if data.strand == "+":
			chromdict[data.strand]["threeprimeend"][data.read_start] += data.number_of_reads
		else:
			chromdict[data.strand]["threeprimeend"][data.read_end-1] += data.number_of_reads

def main(): 
	parser = OptionParser(usage="usage: %prog [options] --gtf=gtf_file.gtf", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("--gtf",dest="input_file",metavar="readdata.gtf",help="type the path to the gtf file data file. By default it expects data from the standard input",default=None)
	files.add_option("-o",dest="output_file",metavar="converted",help="provide a name for an output file. A file extension or strand information is not necessary.",default=None)
	files.add_option("-c", "--chromfile",dest="chromfile",help="Location of the chromosome info file. This file should have two columns: first column is the names of the chromosomes, second column is length of the chromosomes. Default is yeast",metavar="yeast.txt",default=None)
	common = OptionGroup(parser, "sgr output options")
	common.add_option("-t","--type",dest="type",choices=["reads","substitutions","deletions","dropoffrates","startpositions","endpositions","endpositionrates"],help="this tool can generate sgr files for reads, start positions, end positions, substitutions, deletions, reverse transcription drop-off rates (useful for iCLIP data). Please use 'substitutions','deletions','dropoffrates','startpositions','endpositions' to indicate the type of data. Default='reads'",default="reads")
	common.add_option("--min",dest="min",type="int",help="Use this flag to set a minimal read coverage that you want to have reported. This could be useful when you want to calculate drop-off rates only for regions with high coverage",default=0)
	common.add_option("-v", "--verbose",action="store_true",help="to print status messages to a log file",default=False)
	common.add_option("--count",dest="count",action="store_true",help="Takes the numbers in the 'score' column of the GTF file as the total number of reads for each position. Default is 1 for each interval.",default=False)
	common.add_option("--zeros",dest="zeros",action="store_true",help="To include chromosomal positions with zero coverage",default=False)
	common.add_option("--permillion",dest="permillion",action="store_true",help="To normalize the coverage to reads/cDNAs per million",default=False)
	parser.add_option_group(files)
	parser.add_option_group(common)
	(options, args) = parser.parse_args()

	status = sys.stdout
	if options.verbose:
		status = open("pyGTF2sgr_log.txt","w")
	out = getfilename(options.input_file)
	if options.output_file: 
		out = options.output_file
	if options.type == "dropoffrates":
		if options.permillion:
			parser.error("\nYou cannot use the dropoffrates type with permillion normalization\n")
	file_1 = "%s_plus_strand_%s.sgr" % (out,options.type)
	file_2 = "%s_minus_strand_%s.sgr" % (out,options.type)
	outputfiles = [file_1,file_2]
	
	if options.type == "dropoffrates":
		gtf2dropoffrates(options.input_file,options.chromfile,out_files=outputfiles,log=status,score=options.count,zeros=options.zeros,min_cov=options.min)
	elif options.type == "endpositionrates":
		gtf2endpositionrates(options.input_file,options.chromfile,out_files=outputfiles,log=status,score=options.count,zeros=options.zeros,min_cov=options.min)
	else:
		gtf2sgr(options.input_file,options.chromfile,out_files=outputfiles,type=options.type,log=status,score=options.count,zeros=options.zeros,min_cov=options.min,permillion=options.permillion)
		
if __name__ == "__main__":
	main()


#!/usr/bin/python
 
__author__      = "Sander Granneman"
__copyright__   = "Copyright 2017"
__version__     = "0.0.1"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "beta"

import numpy as np
import pandas as pd

class CountTableNormalization():
	def __init__(self):
		self.data = None
		self.datafile = None
		self.genelengths = []
		
	def loadDataFile(self,datafile):
		self.data = pd.read_csv(datafile,index_col=0,header=0,comment="#",sep="\t")
		self.datafile = datafile

	def zscoreNormalization(self):
		return (self.data - self.data.mean(axis=0))/self.data.std(ddof=0)
	
	def __calcualteDEseqScalingfactor(self,counts):
		"""
		Calculate DESeq scaling factor per sample.
		"""
		# masked array to discard inf, -inf, and nan
		ma = np.ma.masked_invalid(counts)
		return np.exp(np.ma.median(ma))

	def DESeqNormalization(self):
		"""
		Normalize by DESeq scaling factor, which is computed as the median of
		the ratio, for each row (gene), of its read count over its geometric
		mean across all samples. Return new counts dataframe.
		Details:
		--------
		http://genomebiology.com/2010/11/10/R106
		Parameters:
		-----------
		- exp_obj: experiment object. Normalized by DESeq scaling factor.
		"""
		df = self.data.copy()
		# log of counts
		lg = df.apply(np.log)
		# per sample: exponential(median(log(counts) - geometric mean))
		sf = lg.sub(lg.mean(axis=1), axis=0).apply(self.__calcualteDEseqScalingfactor, axis=0)
		# apply scaling
		df = df.div(sf, axis=1)
		return df

	def __sf_tmm(self,obs, ref,
				log_ratio_trim=0.3,
				sum_trim=0.05,
				weighting=True,
				a_cutoff=-1e10):
		"""
		Called by `norm_tmm`.
		"""
		if all(obs == ref):
			return 1
	
		obs_sum = obs.sum()
		ref_sum = ref.sum()

		# log ration of expression accounting for library size
		lr = np.log2((obs / obs_sum) / (ref / ref_sum))
		# absolute expression
		ae = (np.log2(obs / obs_sum) + np.log2(ref / ref_sum)) / 2
		# estimated asymptotic variance
		v = (obs_sum - obs) / obs_sum / obs + (ref_sum - ref) / ref_sum / ref
		# create mask
		m = np.isfinite(lr) & np.isfinite(ae) & (ae > a_cutoff)
		# drop the masked values
		lr = lr[m]
		ae = ae[m]
		v = v[m]
		assert len(lr) == len(ae) == len(v)
	
		n = len(lr)
		lo_l = np.floor(n * log_ratio_trim) + 1
		hi_l = n + 1 - lo_l
		lo_s = np.floor(n * sum_trim) + 1
		hi_s = n + 1 - lo_s
		k = ((lr.rank(method="first") >= lo_l) & (lr.rank(method="first") <= hi_l)) \
				& ((ae.rank(method="first") >= lo_s) & (ae.rank(method="first") <= hi_s))

		if weighting:
			return 2**(sum(lr[k] / v[k]) / sum(1 / v[k]))
		else:
			return 2**(lr[k].mean())

	def TMMNormalization(self,ref_col=None,log_ratio_trim=0.3,sum_trim=0.05,weighting=True,a_cutoff=-1e10):
		"""
		Trimmed Mean of M-values (TMM) is the weighted mean of log ratios between
		this test and the reference, after exclusion of the most expressed genes
		and the genes with the largest log ratios.
	
		Parameters:
		-----------
		- exp_obj: experiment object.
		- ref_col: reference column from which to scale others.
		- log_ratio_trim: amount of trim to use on log-ratios.
		- sum_trim: amount of trim to use on combined absolute values.
		- weighting: whether to compute weights.
		- a_cutoff: cutoff of absolute expression values.
		"""
		df = self.data.copy()
		# remove zeros
		nz = df.where(df > 0)
		nz = nz.dropna(how="all").fillna(0)
		# reference column
		if ref_col is None:
			# quantile factors
			sf_q = self.__sf_q(nz)
			ref_col = (abs(sf_q - np.mean(sf_q))).idxmin()
		# try:
		kwargs = {"ref": nz[nz.columns[ref_col]],
				  "log_ratio_trim": log_ratio_trim,
				  "sum_trim": sum_trim,
				  "weighting": weighting,
				  "a_cutoff": a_cutoff}
		# except KeyError:
			# revert back to auto?
		sf_tmm = nz.apply(self.__sf_tmm, **kwargs)
		# apply scaling
		df = df.div(sf_tmm, axis=1)
		return df

	def __sf_q(self,df, q=0.75):
		"""
		Parameters:
		-----------
		- df: zeroed rows removed
		- q: quartile
		"""
		lib_size = df.sum()
		y = df.T.div(lib_size, axis=0).T
		# fill nans with 0
		y = y.dropna(how="all").fillna(0)
		y = y.quantile(q)
		# factors multiple to one
		sf = y.div(np.exp(np.mean(np.log(y))))
		return sf

	def quantileNormalization(self,q=0.75):
		"""
		Ported from edgeR and still needs to be validated. Also, maybe compare
		edgeR method to limma implementation.
	
		Quantile normalization.
	
		Parameters:
		-----------
		- exp_obj: experiment object.
		- q: quantile.
		"""
		df = self.data.copy()
		# remove zeros
		nz = df.where(df > 0)
		nz = nz.dropna(how="all").fillna(0)
		sf_q = self.__sf_q(nz, q)
		# apply scaling
		df = df.div(sf_q, axis=1)
		return df
		
	def FPKMNormalization(self,norm=1000000):
		""" Calculates the FPKM values for each column """
		sumofcollums = self.data.sum(axis=0)
		normdata = self.data.div((sumofcollums/norm),axis=1)
		return normdata.div(self.genelengths,axis=0)
	
	def TPMNormalization(self,norm=1000000):
		""" Calculates Transcripts per million for the raw count data """
		normdata = self.data.div(self.genelengths,axis=0)
		return normdata.div(normdata.sum(axis=0)/norm)


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.5.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyReadCounters.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import os
from optparse import *
from pyCRAC.Classes.Counters import PyCounters
from pyCRAC.Parsers import GTF2

def main(): 
	parser = OptionParser(usage="usage: %prog [options] -f filename", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("-f", "--input_file", dest="filename",help="provide the path to your novo, SAM/BAM or gtf data file. Default is standard input. Make sure to specify the file type of the file you want to have analyzed using the --file_type option!", metavar="FILE",default=None)
	files.add_option("-o", "--output_file", dest="output_file",help="Use this flag to override the standard file names. Do NOT add an extension.", default=None)
	files.add_option("--file_type",dest="file_type",choices=["novo","sam","gtf"],help="use this option to specify the file type (i.e. 'novo','sam' or 'gtf'). This will tell the program which parsers to use for processing the files. Default = 'novo'",default="novo")
	files.add_option("--gtf",dest="annotation_file",metavar="annotation_file.gtf",help="type the path to the gtf annotation file that you want to use",default=None)
	common = OptionGroup(parser, "Common pyCRAC options")
	common.add_option("-a","--annotation",dest="annotation",help="select which annotation (i.e. protein_coding, ncRNA, sRNA, rRNA, tRNA, snoRNA) you would like to focus your search on. Default = all.", metavar="protein_coding",default=None)
	common.add_option("-v","--verbose",dest="verbose",action="store_true",help="prints all the status messages to a file rather than the standard output",default=False)
	common.add_option("--ignorestrand",dest="nostrand",action="store_true",help="To ignore strand information and all reads overlapping with genomic features will be considered sense reads. Useful for analysing ChIP or RIP data",default=False)
	common.add_option("--overlap",dest="overlap",help="sets the number of nucleotides a read has to overlap with a gene before it is considered a hit. Default =  1 nucleotide",type="int",metavar="1",default="1")
	common.add_option("-r","--range",dest="range",type="int",metavar="100",help="allows you to add regions flanking the genomic feature. If you set '-r 50' or '--range=50', then the program will add 50 nucleotides to each feature on each side regardless of whether the GTF file has genes with annotated UTRs.",default=False)
	common.add_option("--rpkm",dest="rpkm",action="store_true",help="to include a column showing reads per kilobase per 1 million mapped reads for each gene. Default is False.",default=False)
	common.add_option("--nucdensities",dest="nucdensities",action="store_true",help="With this flag nucleotide densities are calculated (i.e. the total number of nucleotides overlapping with a feature). Default is False.",default=False)
	common.add_option("--properpairs",dest="properpairs",action="store_true",help="In case you have paired-end data, you can use this flag to analyze only properly paired reads. Default is False",default=False)
	common.add_option("--sense",dest="sense",action="store_true",help="this option instructs pyReadCounters to only consider reads overlapping sense with genomic features. Default is False",default=False)
	common.add_option("--anti_sense",dest="anti_sense",action="store_true",help="this option instructs pyReadCounters to only consider reads overlapping anti-sense with genomic features. Default is False",default=False)
	novo = OptionGroup(parser, "Options for SAM/BAM and Novo files")
	novo.add_option("--force_single_end",dest="force_single_end",action="store_true",help="To consider all reads unpaired, even if it is paired end data.Default is False.",default=False)
	novo.add_option("--hittable",dest="hittable",action="store_true",help="use this flag if you only want to print a hit table for your data",default=False)
	novo.add_option("--gtffile",dest="gtffile",action="store_true",help="use this flag if you only want to print the pyReadCounters gtf file for your data",default=False)
	novo.add_option("--stats",dest="stats",action="store_true",help="use this flag if you only want to print a the statistics showing the complexity of your data",default=False)
	novo.add_option("-s","--sequence",dest="sequence",metavar="intron",choices=["intron","genomic","CDS","exon","5UTR","3UTR"],help="with this option you can select whether you want to count reads overlapping coding or genomic sequence or intron, exon, CDS, 5UTR or 3UTR coordinates. Default = genomic.",default="genomic")
	novo.add_option("--mutations",dest="muts",metavar="delsonly",help="Use this option to only track mutations that are of interest. For CRAC data this is usually deletions (--mutations=delsonly). For PAR-CLIP data this is usually T-C mutations (--mutations=TC). Other options are: do not report any mutations: --mutations=nomuts. Only report specific base mutations, for example only in T's, C's and G's :--mutations=[TCG]. The brackets are essential. Other nucleotide combinations are also possible",default=None)
	novo.add_option("--align_quality", "--mapping_quality", dest="align_quality", metavar="100", type="int", help="with these options you can set the alignment quality (Novoalign) or mapping quality (SAM) threshold. Reads with qualities lower than the threshold will be ignored. Default = 0", default=None)
	novo.add_option("--align_score", dest="align_score", metavar="100", type="int", help="with this option you can set the alignment score threshold. Reads with alignment scores lower than the threshold will be ignored. Default = 0", default=None)
	novo.add_option("--unique",dest="unique",action="store_true",help="with this option reads with multiple alignment locations will be removed. Default = Off",default=False)	 
	novo.add_option("--blocks",dest="blocks",action="store_true",help="with this option reads with the same start and end coordinates on a chromosome will be counted as one cDNA. Default = Off",default=False)
	novo.add_option("-d","--distance",dest="distance",metavar="1000",type="int",help="this option allows you to set the maximum number of base-pairs allowed between two non-overlapping paired reads. Default = 1000",default=1000)
	novo.add_option("--discarded",dest="discarded",metavar="FILE", help="prints the lines from the alignments file that were discarded by the parsers. This file contains reads that were unmapped (NM), of poor quality (i.e. QC) or paired reads that were mapped to different chromosomal locations or were too far apart on the same chromosome. Useful for debugging purposes",default=None)
	novo.add_option("-l","--length",dest="length",metavar="100",type="int",help="to set read length threshold. Default = 1000",default=1000)
	novo.add_option("-m","--max",dest="max_reads",metavar="100000",type="int",help="maximum number of mapped reads that will be analyzed. Default = All",default=float("infinity"))
	novo.add_option("--zip", dest="zip", metavar="FILE", help="use this option to compress all the output files in a single zip file", default=None)
	parser.add_option_group(files)
	parser.add_option_group(common)
	parser.add_option_group(novo)
	(options, args) = parser.parse_args()
	status = sys.stdout
	sense  = True
	anti_sense = True
	annotation = None
	if options.verbose:
		status = open("pyReadCounters_log.txt","w")
	if not options.filename: 
		options.filename = sys.stdin
	if options.muts:
		mutslist = re.compile("\[([A-Za-z]+)\]")
		mutstype = mutslist.findall(options.muts)
		if mutstype: options.muts = [i for i in mutstype[0]]		
	if len(sys.argv) < 1:
		parser.error("usage: %prog -f filename --gtf=yourfavgtffile.gtf [options]. Use -h or --help for options\n")	 
	if options.sense and options.anti_sense:
		parser.error("cannot combine the flags --sense and --anti_sense in a command line\n")
	if options.properpairs and options.force_single_end:
		parser.error("cannot combine the flags --force_single_end and --properpairs in a command line\n")
	if options.sense:
		anti_sense = False
	if options.anti_sense:
		sense = False
	if options.annotation:
		annotation = options.annotation

	### setting which output files to print
	gtf_file   = True
	stats_file = True
	hittable   = True
	if options.hittable:
		gtf_file   = False
		stats_file = False
	if options.gtffile:
		hittable   = False
		stats_file = False
	if options.stats:
		hittable = False
		gtf_file = False
	###

	status.write("parsing GTF annotation file...\n")

	### Making GTF2 parser object and parsing GTF annotation file ###
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.annotation_file,ranges=options.range,transcripts=False,source=annotation)
	###
	
	### creating a PyCounters object
	data = PyCounters(options.filename,gtf,file_type=options.file_type,ignorestrand=options.nostrand,debug=options.discarded)
	###
	
	file_types = ["sam","novo","gtf"]
	if options.file_type in file_types[:2]:			# if the file type is a novo or sam file.
		status.write("processing %s file...\n" % options.file_type)
		data = PyCounters(options.filename,gtf,file_type=options.file_type,ignorestrand=options.nostrand,debug=options.discarded)
		status.write("cataloguing read mapping positions...\n")
		data.countReads(align_quality=options.align_quality,align_score=options.align_score,length=options.length,maximum=options.max_reads,force_single_end=options.force_single_end,distance=options.distance,blocks=options.blocks,unique=options.unique,muts_filter=options.muts,properpair=options.properpairs)
	elif options.file_type == file_types[2]:		# if the file type is a gtf file
		status.write("processing GTF data file...\nNOTE!: flags specific for novo or SAM/BAM files will be ignored!\n")
		data = PyCounters(options.filename,gtf,file_type=options.file_type,ignorestrand=options.nostrand,debug=options.discarded)
		data.dataToDict()
	else:
		parser.error("file type %s is not supported. Please choose from 'sam', 'novo' or 'gtf'" % (options.file_type))
		
	### processing the interval data	
	status.write("counting overlap with genomic features...\n")
	if options.nucdensities:
		data.countNucleotideDensities(min_overlap=options.overlap,sense=sense,anti_sense=anti_sense,sequence=options.sequence)
	else:
		data.countGeneHits(min_overlap=options.overlap,sense=sense,anti_sense=anti_sense,sequence=options.sequence)
	###
	
	if options.discarded:
		data.file_list.append(options.discarded)
	if hittable:
		status.write("printing hittable...\n")
		data.printHitTable(out_file=options.output_file,rpkm=options.rpkm)
	if options.file_type in file_types[:2]:
		if stats_file:
			status.write("printing read statistics...\n")
			data.printFileStats(out_file=options.output_file)
		if gtf_file:
			status.write("printing counter results to GTF files...\n")
			#data.printIntronUTROverlapToGTF(min_overlap=options.overlap,out_file=options.output_file)
			data.printCountResultsToGTF(out_file=options.output_file)
	###
	
	### if requested, compress output files
	if options.zip:					
		import zipfile
		status.write("compressing files...\n")
		file = zipfile.ZipFile(options.zip,mode="w",compression=zipfile.ZIP_DEFLATED)
		for files in data.file_list:
			file.write(files)
		file.close()	
		for files in data.file_list:
			try:
				os.remove(files)
			except OSError:
				pass
	###
						
if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "1.6.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	GTF2.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import numpy as np
import time
import copy
import itertools
from collections import defaultdict
from pyCRAC.Classes.Exceptions import *
from pyCRAC.Classes.NGSFormatWriters import NGSFileWriter
from pyCRAC.Methods import splitter,reverse_complement,splitString
from pyCRAC.Parsers.fasta2dict import *
from pyCRAC.Parsers.tab2dict import *

#--------------------------------------------------- GenomeSeq class ---------------------------------------------------#

class GenomeSeq():
	""" Takes fasta and tab files and converts them into dictionary """
	def __init__(self):
		self.chr_seq = defaultdict(str)
			
	def read_FASTA(self,fasta_file,chromosomes=[]):
		""" Generates a dictionary from a fasta file"""
		self.chr_seq = Fasta2Dict(fasta_file,ids=chromosomes)
			
	def read_TAB(self,tab_file,chromosomes=[]):
		""" Generates a dictionary from a tab file"""
		self.chr_seq = Tab2Dict(tab_file,ids=chromosomes)
			
	def chromosome_list(self):
		""" returns a list of chromosome names"""
		return [keys for keys in sorted(self.chr_seq)]
							
	def chromosome_length(self,chromosome):
		""" returns the length of a chromosome"""
		if chromosome in self.chr_seq:
			return len(self.chr_seq[chromosome])
		else:
			raise LookupError("I could not find %s in the chromosome list\n" % (chromosome))
					
	def sequence(self,chromosome,strand,upstream,downstream):
		"""returns the nucleotide sequence from "chromosome" using upstream and downstream coordinates.Usage: sequence(chromosome,strand,upstream,downstream)"""
		sequence = str()
		try:
			sequence = self.chr_seq[chromosome][upstream:downstream]
		except IndexError:
			if upstream	  < 0 : upstream = 0
			if downstream > len(chromosome): downstream == len(chromosome)
			sequence = self.chr_seq[chromosome][upstream:downstream]
			
		if not sequence:
			raise LookupError("I could not find a sequence for chromosome %s and coordinates %d-%d, please check your input\n" %(chromosome,upstream,downstream))
				
		if strand == "-" or strand == "R":
			return reverse_complement(sequence)
		else:
			return sequence

#--------------------------------------------------- Transcript class ---------------------------------------------------#

class Transcripts():
	"""A class to store transcript information"""
	def __init__(self):
		self.transcripts = dict()
		self.__features = ["exon","CDS","5UTR","3UTR"]
		self.__othersources	 = ["cluster","motif","reads"]
		self.__otherfeatures = ["start_codon","stop_codon"]
					
	def addTranscript(self,transcript_name,start,end,type,annotation,chromosome,strand,gene_name=None,gene_id=None,transcript_id=None):		# gene_name, gene_id and transcript_id are optional features
		"""method to add new transcripts to the self.transcripts dictionary"""
		if transcript_name not in self.transcripts:						# If the transcript dictionary for a particular transcript does not exist, make one and update the key values.
			self.transcripts[transcript_name] = {						# this dictionary is used to store transcript names and transcript information.
				'chromosome'	: chromosome,
				'gene_id'		: gene_id,
				'gene_name'		: gene_name,
				'protein_id'	: str(),
				'strand'		: strand,
				'annotations'	: set(),
				'exon'			: set(),
				'CDS'			: set(),
				'5UTR'			: list(),
				'3UTR'			: list(),
				'start_codon'	: tuple(),
				'stop_codon'	: tuple(),
				'transcript_id' : transcript_id
				}

		self.transcripts[transcript_name]['annotations'].add(annotation)
		if type in self.__features[:2]:									# the features listed in self.__features are commonly found in GTF files. 
			self.transcripts[transcript_name][type].add((start,end))
		elif type in self.__features[2:]:								
			self.transcripts[transcript_name][type].append((start,end))
		if annotation in self.__othersources:							# when processing pyClusterReads, pyMotif or pyReadCounters GTF files
			self.transcripts[transcript_name]['exon'].add((start,end))

		return True

#--------------------------------------------------- Genes class ---------------------------------------------------#

class Genes():
	"""A class to store gene information"""
	def __init__(self):
		counter = 0
		self.__features = ["exon","CDS","5UTR","3UTR"]
		self.__othersources	 = ["cluster","motif","reads"]
		self.genes = dict()
					
	def addGene(self,gene_name,start,end,type,annotation,chromosome,strand,gene_id=None,transcript_name=None):	# gene_id, transcript_name and transcript_id are optional features
		"""method to add new genes to the to the self.genes dictionary"""
		if gene_name not in self.genes:									# If the gene dictionary for a particular gene does not exist, make one and update the key values.
			self.genes[gene_name] = {									# this dictionary is used to store gene names and gene information.									   
				'chromosome'	: chromosome,
				'gene_id'		: gene_id,
				'strand'		: strand,
				'exon'			: set(),
				'CDS'			: set(),
				'5UTR'			: list(),
				'3UTR'			: list(),
				'transcripts'	: set(),
				'annotations'	: set()
				}

		self.genes[gene_name]['annotations'].add(annotation)
		if transcript_name:
			self.genes[gene_name]['transcripts'].add(transcript_name)
		if type in self.__features[:2]:										# the features listed in self.__features are commonly found in GTF files. 
			self.genes[gene_name][type].add((start,end))
		elif type in self.__features[2:]:
			self.genes[gene_name][type].append((start,end))
		if annotation in self.__othersources:								# when processing pyClusterReads, pyMotif or pyReadCounters GTF files
			self.genes[gene_name]['exon'].add((start,end))
		return True

#--------------------------------------------------- GTF2 class ---------------------------------------------------#

class Parse_GTF(GenomeSeq,Genes,Transcripts):
	"""reads the GTF file and stores the info into memory. You need to supply a GTF file name (including path)
	The 'ranges' option allows you to manually set the lengths of the 5' and 3' UTR or flanking sequences."""
	def __init__(self):		 
		"""initialises the class attributes"""
		GenomeSeq.__init__(self)
		Genes.__init__(self)
		Transcripts.__init__(self)
		self.__ranges	 = int()
		self.chromosomes = set()
		self.gtfsources	 = set()
			
	def read_GTF(self,gtf_file,ranges=0,genes=[],source=None,transcripts=True):		# you can set 'ranges' at this stage or later when running methods.
		"""parses the GTF file"""
		self.__ranges = ranges
		with open(gtf_file) as gtf:
			for line in gtf:
				try:
					if line[0] == "#": continue
					Fld = line.strip('\n').split('\t')	  
					chromosome,annotation,type,start,end,strand = Fld[0],Fld[1],Fld[2],int(Fld[3])-1,int(Fld[4]),Fld[6]
					[g_name,g_id,t_name,t_id,g_biotype] = self.getAttributes(Fld[8].strip())
					if g_biotype: annotation = g_biotype							# the more recent GTF files from ENSEMBL have "gene_biotype" in the comments section. Use this rather than column 2 (source)
					if source and source != annotation: continue					# in case you only want to analyse genes associated with specific annotations, such as protein_coding.
					if genes and g_name not in genes: continue						# in case you only want to look at a handfull of genes from the gtf file.
					if g_name == "no_matches": continue								# entries containing "no_matches" can be found in the pyReadCounters output files and these should be ignored.
					self.chromosomes.add(chromosome)
					self.gtfsources.add(annotation)			
					self.addGene(		   
								g_name,												# adding gene names 
								start,
								end,
								type,
								annotation,
								chromosome,
								strand,
								gene_id=g_id,
								transcript_name=t_name,
								)
					if transcripts and t_name:										# transcript positions can also be added if needed				
						self.addTranscript(	   
								t_name,
								start,
								end,
								type,
								annotation,
								chromosome,
								strand,
								gene_name=g_name,
								gene_id=g_id,
								transcript_id=t_id
								)
				except IndexError:
					sys.stderr.write("\nIndexError at line:\n%s\n" % line)
					pass
			
#------------------------------------------------------------------------------ API ------------------------------------------------------------------------------	   

	def getAttributes(self,attributes):
		"""takes the attributes from column 9 in the GTF file and extracts gene_name,transcript_name, etc information"""
		info			= attributes.split(";")
		gene_id			= str()
		transcript_id	= str()
		gene_name		= str()
		transcript_name = str()
		gene_biotype    = str()
		for i in info:
			i = i.strip()
			if i.startswith('gene_id'):
				gene_id = splitter(i)	   
			elif i.startswith('transcript_id'):
				transcript_id = splitter(i)
			elif i.startswith('gene_name'):
				gene_name = splitter(i)
			elif i.startswith('transcript_name'):
				transcript_name = splitter(i)
			elif i.startswith('gene_biotype'):
				gene_biotype = splitter(i)
						
		if not gene_name:
			gene_name = gene_id
		if not transcript_name:
			transcript_name = transcript_id
		return [gene_name,gene_id,transcript_name,transcript_id,gene_biotype]

	def printGeneFlankingCoordinatesAsGTF(self,output=None):
		""" For each gene in the genes dictionary it calculates its extremities based on exon coordinates and prints these out as a gtf file.
		This is useful when using bedtools to look for overlap between intervals and GTF annotation files."""
		if output: out = NGSFileWriter(output)
		out.write("##gff-version 2\n")
		for i in sorted(self.chromosomes):
			for j in sorted(self.chromosomeGeneCoordIterator(i)):
				for annotation in self.annotations(j[2]):
					out.writeGTF(i,annotation,"exon",j[0]-1,j[1],score=".",strand=self.strand(j[2]),frame=".",gene_name=j[2],transcript_name=j[2],gene_id=self.gene2orf(j[2]),transcript_id=self.gene2orf(j[2]),exon_number=None,comments=None)

	def strand(self,gene):
		"""returns the strand (watson or crick: "F" or "R") on which the query gene or ORF is located. Usage: strand("BET1L")"""
		if gene in self.genes:	
			return self.genes[gene]['strand']
		elif gene in self.transcripts:
			return self.transcripts[gene]['strand']
		else:
			raise LookupError("I could not find a strand for gene %s" % gene)
					
	def chromosome(self,gene):
		"""returns the chromosome name on which the query gene or ORF is located. Usage: chromosome("BET1L")"""
		if gene in self.genes:
			return self.genes[gene]['chromosome']
		elif gene in self.transcripts:
			return self.transcripts[gene]['chromosome']
		else:
			raise LookupError("I could not find a chromosome for gene %s" % gene)

	def gene2orf(self,gene):
		"""allows you to convert a gene name to an open reading frame (ORF) name. Usage gene2orf("BET1L")"""
		if gene in self.genes:
			return self.genes[gene]['gene_id']
		elif gene in self.transcripts:
			return self.transcripts[gene]['gene_id']
		else:
			return gene
						
	def orf2gene(self,gene_id):
		"""allows you to convert an open reading frame (ORF) name to a gene name. Usage orf2gene("ENSG00000177951")"""
		result = False
		for x in self.genes:
			if self.genes[x]['gene_id'] == gene_id:
				return x
		return gene_id

	def geneLength(self,gene,type="gene",ranges=0):
		"""returns the length of a gene or transcript. If type is 'gene', it returns the length of a gene_name.
		If type= 'transcript' it returns the length of a transcript_name. """
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		up,down = self.__calcExtremities(gene,gene_type=type,ranges=flank_seq)
		return down-up
	
	def annotations(self,gene):
		"""returns a list of annotations associated with a particular gene or transcript. Requires a gene or transcript name input"""
		if gene in self.genes:
			return self.genes[gene]['annotations']
		elif gene in self.transcripts:
			return self.transcripts[gene]['annotations']
		else:
			return None
					
	def geneTranscriptList(self,gene):
		"""returns a list of (alternatively spliced) transcripts associated with a gene"""
		assert self.genes[gene]['transcripts'], "\n\ncould not find any transcripts for %s\n" % gene
		return self.genes[gene]['transcripts']
			
	def __checkDict(self,gene,dictionary,chromosome=None,annotation=None,strand=None):
		"""checks the genes dictionary to see if it finds any genes that match the specified criteria:
		mapped to 'chromosome'? does it belong to 'annotation'? Is it located on strand 'strand'? 
		The dictionary should be self.genes or self.transcripts """
		results = list()
		if chromosome: results.append(dictionary[gene]['chromosome'] == chromosome)
		if annotation: results.append(annotation in dictionary[gene]['annotations'])
		if strand	 : results.append(dictionary[gene]['strand'] == strand)
		if False in results: 
			return False
		else: 
			return True
			
	def chromosomesTranscriptList(self,chromosome,annotation=None,strand=None):
		"""returns a list of transcripts located on a specified chromosome. One can also select genes with specific annotations or strands.
		Strand should be "+" or "-" and the annotation should be in the gtfsources list.
		Use the pyGetSourcesFromGTF.py script if you want to have a list of all annotations in your GTF file.
		
		Examples: 
		chromosomesTranscriptList("chrXII") = ["A","B","C","D"]
		chromosomesTranscriptList("chrXII",annotation="protein_coding") = ["A","C","D"]
		chromosomesTranscriptList("chrXII",annotation="protein_coding",strand="+") = ["A","D"]
		"""
		if annotation:	
			assert annotation in self.gtfsources, "\n\nthe specified annotation %s does not exist. Choices are:\n%s" % (annotation,"\n".join(self.gtfsources))
		if strand:
			strandoptions = ["+","-"]
			assert strand in strandoptions, "\n\nthe strand %s does not exist\n" % strand
		
		transcriptlist = [i for i in self.transcripts if self.__checkDict(i,self.transcripts,chromosome=chromosome,annotation=annotation,strand=strand)]
		assert transcriptlist, "\n\n Could not find any transcripts on chromosome %s\n" % chromosome
		return transcriptlist
					
	def chromosomesGenesList(self,chromosome,annotation=None,strand=None):
		"""returns a list of genes located on a specified chromosome. One can also select genes with specific annotations or strands.
		Strand should be "+" or "-" and the annotation should be in the gtfsources list.
		Use the pyGetSourcesFromGTF.py script if you want to have a list of all annotations in your GTF file.
		
		Example: 
		chromosomesGenesList("chrXII") = ["A","B","C","D"]
		chromosomesGenesList("chrXII",annotation="protein_coding") = ["A","C","D"]
		chromosomesGenesList("chrXII",annotation="protein_coding",strand="+") = ["A","D"]
		"""
		if annotation:
			assert annotation in self.gtfsources, "\n\nthe specified annotation %s does not exist. Choices are:\n%s" % (annotation,"\n".join(self.gtfsources))
		if strand:
			strandoptions = ["+","-"]
			assert strand in strandoptions, "\n\nthe strand %s does not exist\n" % strand
								
		genelist = [i for i in self.genes if self.__checkDict(i,self.genes,chromosome=chromosome,annotation=annotation,strand=strand)]
		assert genelist, "\n\n Could not find any genes on chromosome %s\n" % chromosome
		return genelist
			
	def __calcExtremities(self,gene,gene_type="gene",ranges=0):
		"""looks at all UTR and exon coordinates, puts them all together and returns the maximum and minimum values in a single tuple.
		if range is set to larger than zero, then UTR coordinates will be ignored"""
		sumlist = list()
		dictionary = dict()
		start = int()
		end	  = int()
		if gene_type == "gene":
			dictionary = self.genes
		elif gene_type == "transcript":
			dictionary = self.transcripts
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)
		
		if ranges:
			sumlist = dictionary[gene]['exon'] | dictionary[gene]['CDS']
			if not sumlist:
				raise NoResultsError("\nNo exon or CDS coordinates available for gene %s. Please correct your GTF annotation file\n" % gene)
			flattenedlist = list(itertools.chain(*sumlist))
			try:
				start = min(flattenedlist) - ranges
				end	  = max(flattenedlist) + ranges
			except ValueError:
				sys.stderr.write("sumlist =  %s\nflattenedlist = %s\n" % (sumlist,flattenedlist))
			if start < 0: start = 0
		else:
			sumlist = dictionary[gene]['exon'] | dictionary[gene]['CDS'] | set(dictionary[gene]['5UTR']) | set(dictionary[gene]['3UTR'])
			flattenedlist = list(itertools.chain(*sumlist))
			try:
				start = min(flattenedlist)
				end	  = max(flattenedlist)
			except ValueError:
				sys.stderr.write("sumlist =  %s\nflattenedlist = %s\n" % (sumlist,flattenedlist))
		if not sumlist:
			raise NoResultsError("\nNo exon or CDS coordinates available for gene %s. Please correct your GTF annotation file\n" % gene)
		return (start,end)
		
	def genomicSequence(self,gene,ranges=0,format=None,split=False):
		"""returns the genomic sequence of transcripts/genes. 
		You need a gene name or ORF name (all UPPERCASE) and you can also set the "ranges" variable (i.e. ranges=50). 
		This will allow you to addd flanking sequences to the genomic sequence. 5UTR or 3UTR coordinates provided in the GTF file
		will automatically be included, unless the ranges value is higher than 0"""
		allowedformats = ["fasta","tab"]
		if format: assert format in allowedformats, "\nthe format you specified is not recognized, please use 'fasta' or 'tab'\n"
		output = str()
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		if gene in self.genes:
			chr_ID = self.genes[gene]['chromosome']
			if self.chr_seq[chr_ID]:
				(up,down) = self.__calcExtremities(gene,gene_type="gene",ranges=flank_seq)
				output = self.chr_seq[chr_ID][up:down]
				if self.genes[gene]['strand'] == "-":
					output = reverse_complement(output)
			else:
				raise NoSequenceError("No sequence information available for %s. Did you load the reference sequence using the read_TAB method?\n" % (gene))	
		elif gene in self.transcripts:
			chr_ID = self.transcripts[gene]['chromosome']
			if self.chr_seq[chr_ID]:
				(up,down) = self.__calcExtremities(gene,gene_type="transcript",ranges=flank_seq)
				output = self.chr_seq[chr_ID][up:down]
				if self.transcripts[gene]['strand'] == "-":
					output = reverse_complement(output)
			else:
				raise NoSequenceError("No sequence information available for %s. Did you load the reference sequence using the read_TAB method?\n" % (gene))			  
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)

		if not format:	
			return output				 
		elif format == "fasta":
			if split:
				return ">%s\n%s" % (gene,splitString(output))
			else:
				return ">%s\n%s" % (gene,output)
		elif format == "tab":				
			return "%s\t%s" % (gene,output)
							
	def codingSequence(self,gene,ranges=0,format=None,split=False):
		"""returns the coding sequence of genes/transcripts. You need a gene name or ORF name (all UPPERCASE) and you can also set the "ranges" variable (i.e. ranges=50). This will allow you to addd flanking sequences to the coding sequence"""
		allowedformats = ["fasta","tab"]
		if format: assert format in allowedformats, "\nthe format you specified is not recognized, please use 'fasta' or 'tab'\n"
		output = str()
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		if gene in self.genes:
			chr_ID = self.genes[gene]['chromosome']
			sequence = str()
			if self.chr_seq[chr_ID]:
				if len(self.genes[gene]['CDS']):
					seq_copy = sorted(copy.deepcopy(self.genes[gene]['CDS']))
					if flank_seq > 0:
						seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
					for coordinates in seq_copy:
						sequence += self.chr_seq[chr_ID][coordinates[0]:coordinates[1]]
					if self.genes[gene]['strand'] == "+":
						output = sequence
					else:
						output = reverse_complement(sequence)
				elif not len(self.genes[gene]['CDS']):
					seq_copy = sorted(copy.deepcopy(self.genes[gene]['exon']))
					if flank_seq > 0:
						seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
					for coordinates in seq_copy:
						sequence += self.chr_seq[chr_ID][coordinates[0]:coordinates[1]]
					if self.genes[gene]['strand'] == "+":
						output = sequence
					else:
						output = reverse_complement(sequence)
			else:
				raise NoSequenceError("No sequence information available for %s. Did you load the reference sequence using the read_TAB method?\n" % (gene)) 
		
		elif gene in self.transcripts:
			chr_ID = self.transcripts[gene]['chromosome']
			sequence = str()
			if self.chr_seq[chr_ID]:
				if self.transcripts[gene]['CDS']:
					seq_copy = sorted(copy.deepcopy(self.transcripts[gene]['CDS']))
					if self.transcripts[gene]['start_codon'] and self.transcripts[gene]['stop_codon']:
						if self.transcripts[gene]['strand'] == "-":
							seq_copy[0]	 = (min(self.transcripts[gene]['stop_codon']), seq_copy[0][1])
							seq_copy[-1] = (seq_copy[-1][0], max(self.transcripts[gene]['start_codon']))
						else:
							seq_copy[0]	 = (min(self.transcripts[gene]['start_codon']), seq_copy[0][1])
							seq_copy[-1] = (seq_copy[-1][0], max(self.transcripts[gene]['stop_codon']))
					if flank_seq > 0:
						seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
					for coordinates in seq_copy:
						sequence += self.chr_seq[chr_ID][coordinates[0]:coordinates[1]]
					if self.transcripts[gene]['strand'] == "+":
						output = sequence
					else:
						output = reverse_complement(sequence)
				elif not self.transcripts[gene]['CDS']:						
					seq_copy = sorted(copy.deepcopy(self.transcripts[gene]['exon']))
					if flank_seq > 0:
						seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
					for coordinates in seq_copy:
						sequence += self.chr_seq[chr_ID][coordinates[0]:coordinates[1]]
					if self.transcripts[gene]['strand'] == "+":
						output = sequence
					else:
						output = reverse_complement(sequence)
			else:
				raise NoSequenceError("No sequence information available for %s. Did you load the reference sequence using the read_TAB method?\n" % (gene))

		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)

		if not format:	
			return output				 
		elif format == "fasta":
			if split:
				return ">%s\n%s" % (gene,splitString(output))
			else:
				return ">%s\n%s" % (gene,output)
		elif format == "tab":				
			return "%s\t%s" % (gene,output)
								 
	def exonCoordinates(self,gene,ranges=0):
		"""Returns a list with tuples of exon coordinates for genes or transcripts. It can have coordinates for multiple exons"""
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		if gene in self.genes:
			if self.genes[gene]['exon']:
				seq_copy = sorted(copy.deepcopy(self.genes[gene]['exon']))
				if flank_seq > 0:
					seq_copy = [(i[0]-flank_seq,i[1]+flank_seq) for i in seq_copy]
				return seq_copy
			else:
				return []
		elif gene in self.transcripts:
			if self.transcripts[gene]['exon']:
				seq_copy = sorted(copy.deepcopy(self.transcripts[gene]['exon']))
				if flank_seq > 0:
					seq_copy = [(i[0]-flank_seq,i[1]+flank_seq) for i in seq_copy]
				return seq_copy
			else:
				return []
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)

	def cdsCoordinates(self,gene,ranges=0):
		"""Returns a list with tuples of CDS coordinates for genes or transcripts. It can have coordinates for multiple exons"""
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		if gene in self.genes:
			if self.genes[gene]['CDS']:
				seq_copy = sorted(copy.deepcopy(self.genes[gene]['CDS']))
				if flank_seq > 0:
					seq_copy = [(i[0]-flank_seq,i[1]+flank_seq) for i in seq_copy]
				return seq_copy
			else:
				return []
		elif gene in self.transcripts:
			if self.transcripts[gene]['CDS']:
				seq_copy = sorted(copy.deepcopy(self.transcripts[gene]['CDS']))
				if flank_seq > 0:
					seq_copy = [(i[0]-flank_seq,i[1]+flank_seq) for i in seq_copy]
				return seq_copy		
			else:
				return []
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)
					
	def fivePrimeSpliceSites(self,gene):
		""" returns five prime splice site coordinates for a gene, if any """
		fiveprimess = list()
		introncoordinates = self.intronCoordinates(gene)
		if introncoordinates:
			for (start,end) in introncoordinates:
				if self.strand(gene) == "+":
					fiveprimess.append(start)
				else:
					fiveprimess.append(end-1)   # 1 has to be subtracted as the end coordinate is a 1-based coordinate!
		return fiveprimess
		
	def threePrimeSpliceSites(self,gene):
		""" returns three prime splice site coordinates for a gene, if any """
		threeprimess = list()
		introncoordinates = self.intronCoordinates(gene)
		if introncoordinates:
			for (start,end) in introncoordinates:
				if self.strand(gene) == "+":
					threeprimess.append(end-1)  # 1 has to be subtracted as the end coordinate is a 1-based coordinate!
				else:
					threeprimess.append(start)
		return threeprimess
		
	def fivePrimeEnd(self,gene,ranges=0):
		""" returns coordinates for the very first nucleotide in the gene. UTRs are included """
		flank_seq = ranges
		if self.__ranges:
			flank_seq = 0		# here you want flank_seq to be ignored if self.__ranges has already been set. Otherwise the two would be added up and longer UTR coordinates would be returned.
		if not gene:
			raise InputError("you forgot to enter the gene/transcript name\n")
		start,end = self.chromosomeCoordinates(gene,ranges=flank_seq)
		if self.strand(gene) == "+":
			return start
		else:
			return end - 1      # 1 has to be subtracted as the end coordinate is a 1-based coordinate!

	def threePrimeEnd(self,gene,ranges=0):
		""" returns coordinates for the very first nucleotide in the gene. UTRs are included """
		flank_seq = ranges
		if self.__ranges:
			flank_seq = 0		# here you want flank_seq to be ignored if self.__ranges has already been set. Otherwise the two would be added up and longer UTR coordinates would be returned.
		if not gene:
			raise InputError("you forgot to enter the gene/transcript name\n")
		start,end = self.chromosomeCoordinates(gene,ranges=flank_seq)
		if self.strand(gene) == "+":
			return end - 1      # 1 has to be subtracted as the end coordinate is a 1-based coordinate!
		else:
			return start
	
	def transcriptionStartSite(self,gene,ranges=0):
		""" returns coordinates for the very first nucleotide for genes that have annotated 5' UTRs """
		flank_seq = ranges
		if self.__ranges:
			flank_seq = 0		# here you want flank_seq to be ignored if self.__ranges has already been set. Otherwise the two would be added up and longer UTR coordinates would be returned.
		if not gene:
			raise InputError("you forgot to enter the gene/transcript name\n")
		(fiveutr,threeutr) = self.utrCoordinates(gene,ranges=flank_seq)
		if fiveutr:
			start,end = fiveutr
			if self.strand(gene) == "+":
				return start
			else:
				return end - 1  # 1 has to be subtracted as the end coordinate is a 1-based coordinate!
		elif not fiveutr:
			start = self.fivePrimeEnd(gene,ranges)
			return start
		else:
			return None
	
	def cdsStart(self,gene):
		""" returns coordinates for the very first nucleotide for genes that have annotated CDS coordinates """
		if not gene:
			raise InputError("you forgot to enter the gene/transcript name\n")
		else:
			coordinates = self.cdsCoordinates(gene)
			flattenedlist = list(itertools.chain(*coordinates))
			if self.strand(gene) == "+":
				return min(flattenedlist)
			else:
				return max(flattenedlist) - 1

	def cdsEnd(self,gene):
		""" returns coordinates for the very first nucleotide for genes that have annotated CDS coordinates """
		if not gene:
			raise InputError("you forgot to enter the gene/transcript name\n")
		else:
			coordinates = self.cdsCoordinates(gene)
			flattenedlist = list(itertools.chain(*coordinates))
			if self.strand(gene) == "+":
				return max(flattenedlist) - 1
			else:
				return min(flattenedlist)
	
	def intronCoordinates(self,gene):
		"""Calculates intron coordinates for genes or transcripts using the exon coordinates present in the 'exon' list. The reason for using the exon
		feature is that not all GTF files have proper CDS coordinates. Splice junctions are usually the same in exon and CDS coordinates. These are 0-based coordinates """
		array = list()
		introns = list()
		if gene in self.genes:
			array = sorted(self.genes[gene]['exon'])
		elif gene in self.transcripts:
			array = sorted(self.transcripts[gene]['exon'])
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)			
		start = int()
		end = int()
		newarray = list()
		for i in array:						# remove exon coordinates that contain positions for start and end coordinates
			if i[-1]-i[0] > 3:
				newarray.append(i)
		newarray.sort()
		if len(newarray) > 1:
			for i in range(0,len(newarray)-1):
				start = max(newarray[i])
				end = min(newarray[i+1])
				if end - start > 4:
					introns.append((start,end))
		return introns
			
	def chromosomeGeneCoordIterator(self,chromosome,annotation=None,strand=None,sequence="genomic",numpy=False,ranges=0):
		"""Returns a list of tuples containing start,end and gene names for a specified chromosome. Format: (start,end,gene_name). 
		The user can also include an annotation or source (from column 2 in the GTF file) to select only genes that, for example, are snRNAs or snoRNA. 
		Make sure that the annotation name is identical to the one in the GTF file. Use the pyGetSourcesFromGTF.py script if you are not sure.
		Ranges allows you to manually set the length of UTR/flanking sequence. If 5UTR and 3UTR coordinates are provided in the GTF file it will try and calculate the extremities of the gene using
		UTR and exon coordinates."""
		gene_list = self.chromosomesGenesList(chromosome,annotation=annotation,strand=strand)
		coordinatesoptions = ["coding", "genomic","intron","exon","CDS","5UTR","3UTR","5end","3end","TSS","CDSstart","CDSend"]
		assert sequence in coordinatesoptions, "Please use the following options for the sequence flag:\n%s\n" % ",".join(coordinatesoptions)
		chromosome_gene_tuple_list = list()
		flank_seq = int()
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		if gene_list:
			for gene in gene_list:
				if sequence == "genomic" or sequence == "coding":
					(up,down) = self.__calcExtremities(gene,gene_type="gene",ranges=flank_seq)
					if up < 0: up = 0
					chromosome_gene_tuple_list.append((up,down-1,gene))
				elif sequence == "CDS":
					cds_coord = self.cdsCoordinates(gene,ranges=flank_seq)
					if cds_coord:
						for (up,down) in cds_coord:
							chromosome_gene_tuple_list.append((up,down-1,gene))
				elif sequence == "exon":
					exon_coord = self.exonCoordinates(gene,ranges=flank_seq)
					if exon_coord:
						for (up,down) in exon_coord:
							chromosome_gene_tuple_list.append((up,down-1,gene))
				elif sequence == "intron":
					intron_coord = self.intronCoordinates(gene)
					if intron_coord:
						for (up,down) in intron_coord:
							chromosome_gene_tuple_list.append((up,down-1,gene))
				elif sequence == "5UTR":
					utrcoord = self.utrCoordinates(gene,ranges=flank_seq)
					if utrcoord and utrcoord[0]:
						(up,down) = utrcoord[0]
						chromosome_gene_tuple_list.append((up,down-1,gene))
				elif sequence == "3UTR":
					utrcoord = self.utrCoordinates(gene,ranges=flank_seq)
					if utrcoord and utrcoord[1]:
						(up,down) = utrcoord[1]
						chromosome_gene_tuple_list.append((up,down-1,gene))
				elif sequence == "5ss":
					fiveprimess = self.fivePrimeSpliceSites(gene)
					if fiveprimess:
						for coord in fiveprimess:
							chromosome_gene_tuple_list.append((coord-flank_seq,coord+flank_seq,gene))
				elif sequence == "3ss":
					threeprimess = self.threePrimeSpliceSites(gene)
					if threeprimess:
						for coord in threeprimess:
							chromosome_gene_tuple_list.append((coord-flank_seq,coord+flank_seq,gene))
				elif sequence == "5end" or sequence == "TSS":
					fiveprimeend = self.fivePrimeEnd(gene)
					chromosome_gene_tuple_list.append((fiveprimeend-flank_seq,fiveprimeend+flank_seq,gene))
				elif sequence == "3end":
					threeprimeend = self.threePrimeEnd(gene)
					chromosome_gene_tuple_list.append((threeprimeend-flank_seq,threeprimeend+flank_seq,gene))
				elif sequence == "CDSstart":
					if self.genes[gene]['CDS']:
						fiveprimeend= self.cdsStart(gene)
						chromosome_gene_tuple_list.append((fiveprimeend-flank_seq,fiveprimeend+flank_seq,gene))
				elif sequence == "CDSend":
					if self.genes[gene]['CDS']:
						threeprimeend = self.cdsEnd(gene)
						chromosome_gene_tuple_list.append((threeprimeend-flank_seq,threeprimeend+flank_seq,gene))
				else:
					raise InputError("I did not recognize the %s option. Please double check your input\n" % sequence)
		if numpy:
			return np.array(chromosome_gene_tuple_list, dtype='i,i,U100')	# gene_name length of 100 characters is the absolute maximum
		else:
			return chromosome_gene_tuple_list
	
	def geneIterCoordinates(self,gene,coordinates="genomic",ranges=0,output="numpy"):
		""" Returns chromosomal positions for a gene as a a numpy, set or python array. Very useful for making pileups!
		By default it returns numpy arrays but you can tell it to return standard python lists or set() lists by stipulating output="list" or output="set"
		as an argument when using this method. Coordinate input choices: genomic, coding,intron, exon, CDS, 5UTR, 3UTR (untranslated regions), 3ss, 5ss (splice sites)
		, TSS (transcription start sites, (if any)), 5end and 3end (5' and 3' end coordinates, respectively). 
		If you choose genomic or coding sequence coordinates, then setting ranges will add extra coordinates upstream and downstream of the sequences. 
		When choosing exon, CDS or intron coordinates, ranges adds extra coordinates to each individual exon, CDS or intron. """
		coordinatesoptions = ["genomic","coding","intron","exon","CDS","5UTR","3UTR","5ss","3ss","TSS","5end","3end","CDSstart","CDSend"]
		assert coordinates in coordinatesoptions, "\n\nCould not find coordinates for %s. Please choose from the following options:\n%s\n" % (coordinates,"\n".join(coordinatesoptions))
		if not gene:
			raise InputError("you forgot to enter the gene/transcript name\nCould there be an error in your GTF annotation file?")
		if coordinates == "genomic":																# returns genomic sequence coordinates. Contain intronic sequences
			coord = self.chromosomeCoordinates(gene,ranges=ranges)
			if output == "numpy":
				return np.arange(coord[0],coord[1],dtype=int)
			elif output == "set":
				return set(range(coord[0],coord[1]))
			elif output == "list":
				return list(range(coord[0],coord[1]))
		elif coordinates == "coding":																# returns coding sequence coordinates. Introns have been removed
			cds_array = list()
			coding_seq_coord = self.codingSequenceCoordinates(gene,ranges=ranges)
			if coding_seq_coord:
				for coord in coding_seq_coord:
					cds_array.extend(list(range(coord[0],coord[1])))
				if output == "numpy":
					return np.array(cds_array,dtype=int)
				elif output == "set":
					return set(cds_array)
				elif output == "list":
					return cds_array
			else:
				return []								
		elif coordinates == "intron":
			intron_array = list()
			intron_coord = self.intronCoordinates(gene)
			if intron_coord:
				for coord in intron_coord:
					if output == "numpy":
						intron_array.append(np.arange(coord[0]-ranges,coord[1]+ranges))				#NOTE!!! ranges adds extra coordinates upstream and downstream of each intron, not entire gene!
					elif output == "set":
						intron_array.append(set(range(coord[0]-ranges,coord[1]+ranges)))
					elif output == "list":
						intron_array.append(list(range(coord[0]-ranges,coord[1]+ranges)))
				return intron_array
			else:
				return []						
		elif coordinates == "exon":
			exon_array = list()
			exon_coord = self.exonCoordinates(gene)
			if exon_coord:
				for coord in exon_coord:
					if output == "numpy":
						exon_array.append(np.arange(coord[0]-ranges,coord[1]+ranges))				#NOTE!!! ranges adds extra coordinates upstream and downstream of each exon, not entire gene!
					elif output == "set":
						exon_array.append(set(range(coord[0]-flank_seq,coord[1]+ranges)))
					else:
						exon_array.append(list(range(coord[0]-ranges,coord[1]+ranges)))
				return exon_array
			else:
				return []						
		elif coordinates == "CDS":
			CDS_array = list()
			CDS_coord = self.cdsCoordinates(gene)
			if CDS_coord:
				for coord in CDS_coord:
					if output == "numpy":
						CDS_array.append(np.arange(coord[0]-ranges,coord[1]+ranges))				#NOTE!!! ranges adds extra coordinates upstream and downstream of each CDS, not entire gene!
					elif output == "set":
						CDS_array.append(set(range(coord[0]-ranges,coord[1]+ranges)))
					else:
						CDS_array.append(list(range(coord[0]-ranges,coord[1]+ranges)))
				return CDS_array
			else:
				return []			
		elif re.search("UTR",coordinates,re.I):
				UTR_coord = self.utrCoordinates(gene,ranges=ranges)
				if coordinates.upper() == "5UTR":
					fiveUTR_coord = UTR_coord[0]
					if fiveUTR_coord:
						if output == "numpy":
							return np.arange(fiveUTR_coord[0],fiveUTR_coord[1])
						elif output == "set":
							return set(range(fiveUTR_coord[0],fiveUTR_coord[1]))
						else:
							return list(range(fiveUTR_coord[0],fiveUTR_coord[1]))
					else:
						return []
								
				elif coordinates.upper() == "3UTR":
					threeUTR_coord = UTR_coord[1]
					if threeUTR_coord:
						if output == "numpy":
							return np.arange(threeUTR_coord[0],threeUTR_coord[1])
						elif output == "set":
							return set(range(threeUTR_coord[0],threeUTR_coord[1]))
						else:
							return list(range(threeUTR_coord[0],threeUTR_coord[1]))
					else:
						return [] 
		elif coordinates in coordinatesoptions[6:]:			# in case it is a single value; "5ss","3ss","TSS","5end" and"3end"
			positions = list()
			if coordinates == "5ss":
				fiveprimess = self.fivePrimeSpliceSites(gene)
				if fiveprimess:
					for coord in fiveprimess:
						positions.append(list(range(coord-ranges,coord+ranges+1)))
			elif coordinates == "3ss":
				threeprimess = self.threePrimeSpliceSites(gene)
				if threeprimess:
					for coord in threeprimess:
						positions.append(list(range(coord-ranges,coord+ranges+1)))
			elif coordinates == "5end":
				fiveprimeend = self.fivePrimeEnd(gene)
				positions = list(range(fiveprimeend-ranges,fiveprimeend+ranges+1))
			elif coordinates == "TSS":
				TSS = self.transcriptionStartSite(gene)
				if TSS:
					positions = list(range(TSS-ranges,TSS+ranges+1))
			elif coordinates == "3end":
				threeprimeend = self.threePrimeEnd(gene)
				positions = list(range(threeprimeend-ranges,threeprimeend+ranges+1))
			elif coordinates == "CDSstart":
				fiveprimeend= self.cdsStart(gene)
				if fiveprimeend:
					positions = list(range(fiveprimeend-ranges,fiveprimeend+ranges+1))
			elif coordinates == "CDSend":
				threeprimeend = self.cdsEnd(gene)
				if threeprimeend:
					positions = list(range(threeprimeend-ranges,threeprimeend+ranges+1))
			else:
				sys.stderr.write("\n\nCould not find coordinates for %s. Please choose from the following options:\n%s\n" % (coordinates,"\n".join(coordinatesoptions)))
			if positions:
				if output == "numpy": 
					return np.array(positions)
				if output == "set":	  
					return set(positions)
				else:
					return positions
			else:
				return []
		else:
			raise InputError("I did not recognize the %s option. Please double check your --sequence input\n" % coordinates)

	def utrCoordinates(self,gene,ranges=0):
		""" This function calculates the UTR coordinates using the exon and CDS coordinates in 
		the a Parse_GTF dictionary object if UTR coordinates were not defined in the GTF file.
		If a range has been set then UTR coordinates in the GTF file will be ignored."""
		dictionary = dict()
		if gene in self.genes:
			dictionary = self.genes
		elif gene in self.transcripts:
			dictionary = self.transcripts
		else:
			sys.stderr.write("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file\n" % gene)
			return None
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		def calc5UTR(gene,dictionary):
			""" calculates 5'UTR coordinates """
			if not flank_seq:
				if dictionary[gene]['5UTR']:
					return dictionary[gene]['5UTR'][0]
				elif not dictionary[gene]['exon']:
					return ()
				elif dictionary[gene]['exon'] and dictionary[gene]['CDS']:
					five_UTR	= tuple()
					listofexons = list(itertools.chain(*dictionary[gene]['exon']))
					listofcds	= list(itertools.chain(*dictionary[gene]['CDS']))
					maxlistofexons = max(listofexons)
					minlistofexons = min(listofexons)
					maxlistofcds  = max(listofcds)
					minlistofcds  = min(listofcds)
					if dictionary[gene]['strand'] == "+":
						if minlistofexons < minlistofcds:
							upstream   = minlistofexons
							downstream = minlistofcds - 1
							five_UTR   = (upstream,downstream)
				 
					elif dictionary[gene]['strand'] == "-":
						if maxlistofexons > maxlistofcds:
							upstream   = maxlistofcds
							downstream = maxlistofexons
							five_UTR   = (upstream,downstream)
					return five_UTR
			elif flank_seq:
				if dictionary[gene]['CDS']:
					five_UTR  = tuple()
					listofcds = list(itertools.chain(*dictionary[gene]['CDS']))
					maxlistofcds  = max(listofcds)
					minlistofcds  = min(listofcds)
					if dictionary[gene]['strand'] == "+":
						downstream = minlistofcds
						upstream   = downstream - flank_seq
						five_UTR   = (upstream,downstream)
					elif dictionary[gene]['strand'] == "-":
						upstream   = maxlistofcds
						downstream = upstream + flank_seq
						five_UTR   = (upstream,downstream)
					return five_UTR
				if not dictionary[gene]['CDS'] and dictionary[gene]['exon']:
					five_UTR	= tuple()
					listofexons = list(itertools.chain(*dictionary[gene]['exon']))
					maxlistofexons = max(listofexons)
					minlistofexons = min(listofexons)
					if dictionary[gene]['strand'] == "+":
						downstream = minlistofexons
						upstream   = downstream - flank_seq
						five_UTR   = (upstream,downstream)
					elif dictionary[gene]['strand'] == "-":
						upstream   = maxlistofexons
						downstream = upstream + flank_seq
						five_UTR   = (upstream,downstream)
					return five_UTR
			else:
				return ()

		def calc3UTR(gene,dictionary):
			""" calculates 3'UTR coordinates """ 
			if not flank_seq:
				if dictionary[gene]['3UTR'] :
					return dictionary[gene]['3UTR'][0]
				elif not dictionary[gene]['exon']:
					return ()
				elif dictionary[gene]['exon'] and dictionary[gene]['CDS']:
					three_UTR	= tuple()
					listofexons = list(itertools.chain(*dictionary[gene]['exon']))
					listofcds	= list(itertools.chain(*dictionary[gene]['CDS']))	 
					maxlistofexons = max(listofexons)
					minlistofexons = min(listofexons)
					maxlistofcds  = max(listofcds)
					minlistofcds  = min(listofcds)
					if dictionary[gene]['strand'] == "+":		
						if maxlistofexons > maxlistofcds:
							upstream   = maxlistofcds
							downstream = maxlistofexons
							three_UTR  = (upstream,downstream)
						else:
							three_UTR  = ()
					elif dictionary[gene]['strand'] == "-":
						if minlistofexons < minlistofcds:
							upstream   = minlistofexons
							downstream = minlistofcds
							three_UTR  = (upstream,downstream)
						else:
							three_UTR  = ()
					return three_UTR	
			elif flank_seq > 0:
				if dictionary[gene]['CDS']:
					three_UTR = tuple()
					listofcds = list(itertools.chain(*dictionary[gene]['CDS']))
					maxlistofcds  = max(listofcds)
					minlistofcds  = min(listofcds)
					if dictionary[gene]['strand'] == "+":		
						upstream   = maxlistofcds
						downstream = upstream + flank_seq
						three_UTR  = (upstream,downstream)
					elif dictionary[gene]['strand'] == "-":
						downstream = minlistofcds
						upstream   = downstream - flank_seq
						three_UTR  = (upstream,downstream)
					return three_UTR
					
				elif not dictionary[gene]['CDS'] and dictionary[gene]['exon']:
					three_UTR	= tuple()
					listofexons = list(itertools.chain(*dictionary[gene]['exon']))
					maxlistofexons = max(listofexons)
					minlistofexons = min(listofexons)
					if dictionary[gene]['strand'] == "+":
						upstream   = maxlistofexons
						downstream = upstream + flank_seq
						three_UTR  = (upstream,downstream)
					elif dictionary[gene]['strand'] == "-":
						downstream = minlistofexons
						upstream   = downstream - flank_seq
						three_UTR  = (upstream,downstream)
					return three_UTR
			else:
				return ()
						
		return [calc5UTR(gene,dictionary),calc3UTR(gene,dictionary)]
			
	def codingSequenceCoordinates(self,gene,ranges=0):
		"""returns a list with tuples of coding sequence (CDS) coordinates for genes or transcripts. It can have coordinates for
		multiple CDSs. Returns a list of exon coordinates if the gene or transcript has no CDS coordinates"""
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
					
		if gene in self.genes:
			if self.genes[gene]['CDS']:
				seq_copy = sorted(copy.deepcopy(self.genes[gene]['CDS']))
				if flank_seq > 0:
					seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
					seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
				return seq_copy
			else:
				seq_copy = sorted(copy.deepcopy(self.genes[gene]['exon']))
				if flank_seq > 0:
					seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
					seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
				return seq_copy						
		elif gene in self.transcripts:
			if self.transcripts[gene]['CDS']:
				seq_copy = sorted(copy.deepcopy(self.transcripts[gene]['CDS']))
				if self.transcripts[gene]['start_codon'] and self.transcripts[gene]['stop_codon']:
					if self.transcripts[gene]['strand'] == "-":
						seq_copy[0]	 = (min(self.transcripts[gene]['stop_codon']), seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], max(self.transcripts[gene]['start_codon']))
					elif self.transcripts[gene]['strand'] == "+":
						seq_copy[0]	 = (min(self.transcripts[gene]['start_codon']), seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], max(self.transcripts[gene]['stop_codon']))
				if flank_seq > 0:
						seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
						seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
				return seq_copy
			else:
				seq_copy = sorted(copy.deepcopy(self.transcripts[gene]['exon']))
				if flank_seq > 0:
					seq_copy[0]	 = (seq_copy[0][0] - flank_seq, seq_copy[0][1])
					seq_copy[-1] = (seq_copy[-1][0], seq_copy[-1][1] + flank_seq)
				return seq_copy			
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)

	def intronSequences(self,gene,format=None,split=False):
		"""returns a list with the intron sequences for a gene or ORF of interest. Returns an empty list if the gene has no introns. Usage: intron("RPL7A")"""
		allowedformats = ["fasta","tab"]
		if format: assert format in allowedformats, "\nthe format you specified is not recognized, please use 'gb', 'fasta' or 'tab'\n"
		introns = list()
		out		= str()
		chr_ID	= str()
		strand	= str()
		if gene in self.transcripts:  
			chr_ID = self.transcripts[gene]['chromosome']
			strand = self.transcripts[gene]['strand']
		elif gene in self.genes:	  
			chr_ID = self.genes[gene]['chromosome']
			strand = self.genes[gene]['strand']
		else:
			raise LookupError("Could not find %s in the genes or transcript dictionary. Note that gene and transcript names must be exactly the same as in the GTF file" % gene)
		if not self.chr_seq[chr_ID]:
			raise NoSequenceError("No intronic sequences available. Did you forgot to load the genomic reference sequence? Use the read_TAB method to load the fasta file containing the genome\n")
		intron_coordinates = sorted(self.intronCoordinates(gene))
		if intron_coordinates:
			for coord in intron_coordinates:
				intron_up	= coord[0]
				intron_down = coord[1]
				if strand  == "+":
					intron = self.chr_seq[chr_ID][intron_up-1:intron_down+1]
					introns.append(intron)
				elif strand	 == "-":
					intron = self.chr_seq[chr_ID][intron_up-1:intron_down+1]
					introns.append(reverse_complement(intron))
		else:
			return None

		if not format:	
			return introns				
		elif format == "fasta":
			for nr,intron in enumerate(introns,start=1):
				if split:
					out += ">%s_intron_%s\n%s\n" % (gene,nr,splitString(intron))
				else:
					out += ">%s_intron_%s\n%s\n" % (gene,nr,intron)
			return out
		elif format == "tab":				
			for intron in introns:
				out += "%s\t%s\n" % (gene,intron)
		return out

	def chromosomeCoordinates(self,gene,ranges=0):
		"""	 returns a tuple of chromosome coordinates for a genes or ORFS. Usage: chromosomeCoordinates("RRP9")"""
		if self.__ranges: 
			flank_seq = self.__ranges
		else:
			flank_seq = ranges
		if gene in self.genes:
			return self.__calcExtremities(gene,gene_type="gene",ranges=flank_seq)
		elif gene in self.transcripts:
			return self.__calcExtremities(gene,gene_type="transcript",ranges=flank_seq)
		else:
			raise LookupError("could not find chromosome coordinates for this %s\n" % (gene))



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.4"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "beta"

##################################################################################
#
#	pybed2GTF.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import time
from optparse import *
from collections import defaultdict
from pyCRAC.Methods import numpy_overlap
from pyCRAC.Parsers import GTF2

def processBed(gtf,datafile,outfile):
	""" A function that reads a bed file and converts it to GTF with features overlapping each interval """
	genedict = defaultdict(list)
	outfile.write("##gff-version 2\n# generated by pybed2GTF.py, %s\n# %s\n" % (time.ctime(),' '.join(sys.argv))) 
	if not datafile:
		infile = sys.stdin
	else:	
		infile = open(datafile,"r")
	for chromosome in gtf.chromosomes:
		genedict[chromosome] = gtf.chromosomeGeneCoordIterator(chromosome,numpy=True)
	for line in infile:
		if line.startswith("#"): continue
		if line.startswith("track"): continue
		Fld = line.strip().split("\t")
		try:
			chromosome = Fld[0]
			start	   = int(Fld[1])
			end		   = int(Fld[2])
			comment	   = Fld[3]
			numberofreads = Fld[4]
			strand	   = Fld[5]
			genes =  numpy_overlap(genedict[chromosome],start,end,overlap=1)
			if genes:
				for gene in genes:
					if gtf.strand(gene) != strand:
						continue
					else:
						outfile.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tgene_name \"%s\"; transcript_name \"%s\";\n" % (chromosome,"interval",comment,start+1,end+1,numberofreads,strand,".",gene,gtf.gene2orf(gene)))	 # note that bed files are 0-based, whereas GTF files are 1-based
			else:
				gene = "no_matches"
				outfile.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\tgene_name \"%s\"; transcript_name \"%s\";\n" % (chromosome,"interval",comment,start+1,end+1,numberofreads,strand,".",gene,gtf.gene2orf(gene)))
		except:
			sys.stderr.write("IndexError at line:\n%s\n" % (line))
			pass	
def main():
	parser = OptionParser(usage="usage: %prog [options] --bed=myfile.bed --gtf=yeast.gtf -o myfile.gtf", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("--bed",dest="bed_file",metavar="Yourfavoritebed.bed",help="type the path to the bed file that you want to convert. Default is standard input",default=None)
	files.add_option("--gtf",dest="gtf_file",metavar="annotations.gtf",help="type the path to the gtf file that you want to use as annotation file. Default is yeast",default=None)
	files.add_option("-o","--outfile",dest="outfile",help="type the name and path of the file you want to write the output to. Default is standard output",default=None)
	parser.add_option_group(files)
	(options, args) = parser.parse_args()
	outfile = sys.stdout
	if options.outfile:
		outfile = open(options.outfile,"w")
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.gtf_file)
	
	processBed(gtf,options.bed_file,outfile)


if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2018"
__version__		= "0.3.4"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	Methods
#
#	CONTAINS USEFUL BITS OF CODE FOR NGS DATA PROCESSING
#
#
#	Copyright (c) Sander Granneman 2018
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import numpy as np
import sys
import os
import io
import re
import six
import array

try:
	file_types = (file, io.IOBase)

except NameError:
	file_types = (io.IOBase,)

#------------------------------------------------------------------------------------------------------------
# Numpy array conversion methods
#------------------------------------------------------------------------------------------------------------

def contigousarray2Intervals(a):
	""" Converts lists and numpy arrays containing consecutive numbers to intervals.
	For example:
	>>> a = np.array([1,3,4,5,6,7,8,10,11,12,13,14,17,18,19])
	>>> intervallist = contigousarray2Intervals(a)
	>>> print(intervallist)
	[(1, 1),(3, 8), (10, 14), (17, 19)]

	"""
	start = None
	end	  = None
	intervallist = list()
	for nr,i in enumerate(list(a)):
		if not start:
			start = i
		try:
			if a[nr+1] == i + 1:
				continue
			elif a[nr+1] > i+1:
				end = i
				intervallist.append((start,end))
				start = a[nr+1]
		except IndexError:
			if start:
				end = i
				intervallist.append((start,end))
			break
	return intervallist

def intervalsFromClosePositions(positions,mindistance=5):
    """ Merges positions that are within a specific distance
    to their neighbours:
    [1,3,5,15,20,30,35,36,37,69,70,80,90,91]
    should become:
    [(1, 5), (15, 20), (30, 37), (69, 70), (80, 80), (90, 91)]

    """
    start = None
    end	  = None
    intervallist = list()
    for nr,i in enumerate(sortedl(list(positions))):
        if not start:
            start = i
        try:
            if positions[nr+1] - positions[nr] <= mindistance:
                continue
            elif positions[nr+1] - positions[nr] > mindistance:
                end = i
                intervallist.append((start,end))
                start = positions[nr+1]
        except IndexError:
            if start:
                end = i
                intervallist.append((start,end))
            break
    return intervallist

def steparray2Intervals(a):
	""" Converts step arrays into a list containing tuples of intervals and coverage.
	Useful for making bed or bedGraph files. The first two values of the tuple are genomic start and end positions, respectively.
	The third value in the tuple is the total number of reads covering the start and end positions.

	For example:
	>>> a = np.array([0,0,0,0,1,1,2,2,3,3,3,4,3,3,3,2,1,1,1,0,0,0,0,1,1,1,1,2,2,2,3,3,3,4,5,6,4,4,4,4,4,4,3,3,3,3,2,2,1,1,1,0,0,0,0,0,1,1,1,0,0,0,0])
	>>> intervallist = steparray2Intervals(a)
	>>> print(intervallist)
	[(4, 5, 1), (6, 7, 2), (8, 10, 3), (11, 11, 4), (12, 14, 3), (15, 15, 2), (16, 18, 1), (19, 22, 0), (23, 26, 1), (27, 29, 2),
	(30, 32, 3), (33, 33, 4), (34, 34, 5), (35, 35, 6), (36, 41, 4), (42, 45, 3), (46, 47, 2), (48, 50, 1), (51, 55, 0), (56, 58, 1)]

	Bedgraph files are zero-based so the first coordinate should always start with a zero.
	"""
	start = 0
	intervallist = list()
	for i in np.where(np.diff(a)!=0)[0]:
		if not start:
			start = i
		else:
			intervallist.append((start,i,int(a[i])))
			start = i
	return intervallist

def mergeShortIntervals(intervals,mindistance=5):
	""" Merges intervals very close to each other with their neighbors.
	[(62352.0, 62450.0),(62452.0, 62491.0),(71760.0, 71806.0),(71808.0, 71820.0)]
	should become:
	(62352.0, 62491.0),(71760.0, 71820.0)

	"""
	newintervals = list()
	while intervals:
		first = intervals.pop(0)
		try:
			second = intervals.pop(0)
		except:
			break
		if second[0]-first[1]<= mindistance:
			intervals.insert(0,(first[0],second[1]))
		else:
			newintervals.append(first) # only put the first interval
			intervals.insert(0,second) # put the second back and start again
	return newintervals

def flattenListofArrays(array):
	""" Takes a list of arrays or tuples and flattens them into a single list """
	return [item for sublist in array for item in sublist]

def mergeIntervalCoordinates(a,b):
	""" merges overlapping interval coordinates. Example:
	a = [(3218948, 3219354),(3215871, 3217986)]
	b = [(3218953, 3219006),(3219170, 3219216)]

	mergeIntervalCoordinates(a,b) returns:

	[(3215871, 3217986),(3218948, 3218952),(3219007, 3219169),(3219217, 3219354)]

	"""
	newcoord = list()
	for i in a:
		if i:
			coords = np.arange(i[0],i[1]+1)
			for j in b:
				index = np.where(np.in1d(coords,np.arange(j[0],j[1]+1)))[0]
				coords = np.delete(coords,index)
			newcoord.extend(contigousarray2Intervals(coords))
	newcoord.sort()
	return newcoord

#------------------------------------------------------------------------------------------------------------
# file indexing tools
#------------------------------------------------------------------------------------------------------------

"""
	How to use these indexing tools. Example:
	>>> # make an index file:
	>>> makeBinaryIndexFile(fn)
	>>> findex = readFileIndex(fn)	# this assumes an index file already exists, otherwise it will make it
	>>> f = open(fn, 'rb')
	>>> position = 3				# position in file you want to go to is line 3
	>>> print(readline(position, f, findex)) # prints line 3
	2-micron	reads	exon	2903	2941	8	+	.	gene_id "INT_0_3,R0020C"; gene_name "INT_0_3,REP1";
	>>> seek(position,f,index)		# does the same
	>>> print(next(f)
	2-micron	reads	exon	2903	2941	8	+	.	gene_id "INT_0_3,R0020C"; gene_name "INT_0_3,REP1";

"""
def reader(f):
	""" function that reads the file in order to make an index """
	blocksize = 1024 * 1024
	blockstart = 0
	while True:
		block = f.read(blocksize)
		if not block: break
		inblock = 0
		while True:
			nextnl = block.find(b'\n', inblock)
			if nextnl < 0:
				blockstart += len(block)
				break
			yield nextnl + blockstart
			inblock = nextnl + 1

def makeBinaryIndexFile(fn):
	""" makes a binary index of a file """
	with open(fn, 'rb') as f:
		# result format: x[0] is tot # of lines,
		# x[N] is byte offset of END of line N (1+)
		result = array.array('L', [0])
		result.extend(reader(f))
		result[0] = len(result) - 1
	with open(fn + '.indx', 'wb') as p:
		result.tofile(p)

def readline(n,f,findex):
	""" seeks to a beginning of a line at 'n' in file 'f' using index 'findex' and returns the line """
	f.seek(findex[n] + 1)
	bytes = f.read(findex[n+1] - findex[n])
	return bytes.decode('utf8')

def seek(linenumber,file,index):
	""" jumps to position in file at linenumber using index """
	file.seek(index[linenumber])
	six.next(file)

def readFileIndex(indexfile):
	""" reads the file index and stores it into an object """
	with open("%s.indx" % indexfile,'rb') as file:
		findex = array.array('l')
		findex.fromfile(file, 1)
		findex.fromfile(file, findex[0])
		findex[0] = -1
	return findex

#------------------------------------------------------------------------------------------------------------

def processChromFile(file,ignore=[]):
	""" Takes a file containing chromosome names and chromosome lengths and returns it as a dictionary.
	The keys in the dictionary are the chromosome names, the values are the chromosome lengths.
	The chromosome info file should be a tab delimited file formatted as follows:
	#chromosome length
	chrI	23452453
	chrII	34244244
	etc...
	"""
	chromdict = dict()
	for line in file:
		if line.startswith("#"): continue
		Fld = line.strip().split()
		chromosome = Fld[0]
		if ignore and chromosome in ignore:
			continue
		else:
			chromdict[Fld[0]] = int(Fld[1])
	return chromdict

def splitData(datafile, directory="./", max_open_filehandles=100):
	""" generates a tmp file for each chromosome in the data using the tempfile module. NOTE that these need to be removed manually by using the unlink function.
	for example:
	temp = tempfile.NamedTemporaryFile(prefix='%s_' % chr,dir='./',suffix='.tmp', delete=False)
	temp.unlink(temp.name)

	By attaching a path to the directory variable you can direct the files to a specific directory.
	Returns a dictionary where the keys are the chromosome names and the values are the file names.
	"""
	import tempfile
	from collections import deque

	# This dictionary contains the complete map of chromosome names to temp file handles (open or closed)
	chromosomes = dict()

	# This dictionary contains a map of chromosome names to open file handles only
	open_tempfiles = dict()

	# This deque represents a FIFO queue to keep track of the order in which temp files were opened. When the maximum
	# number of open filehandles is reached, the oldest file is closed
	open_chromosome_queue = deque()

	with open(datafile, "r") as data:		 # going through the data line by line
		for line in data:
			if line.startswith("track"):
				continue
			if line.startswith("#"):
				continue
			chr = line.strip().split("\t")[0]
			if chr not in open_tempfiles and chr in chromosomes:
				# open a new filehandle using an existing tempfile, closing an old filehandle if necessary
				open_chromosome_queue.append(chr)
				if len(open_chromosome_queue) > max_open_filehandles:
					chromosome_file_to_close = open_chromosome_queue.popleft()
					open_tempfiles[chromosome_file_to_close].close()
					del open_tempfiles[chromosome_file_to_close]
				temp_filepath = os.path.join(directory, chromosomes[chr].name)
				open_tempfiles[chr] = open(temp_filepath, 'w')

			elif chr not in open_tempfiles and chr not in chromosomes:
				# create a new filehandle with a new tempfile, closing an old filehandle if necessary
				temp = tempfile.NamedTemporaryFile(mode='w',prefix='%s_' % chr, dir=directory, suffix='.tmp', delete=False)
				chromosomes[chr] = temp
				open_chromosome_queue.append(chr)
				if len(open_chromosome_queue) > max_open_filehandles:
					chromosome_file_to_close = open_chromosome_queue.popleft()
					open_tempfiles[chromosome_file_to_close].close()
					del open_tempfiles[chromosome_file_to_close]
				open_tempfiles[chr] = temp

			open_tempfiles[chr].write(line)

			# optional sanity check
			if (set(open_chromosome_queue) != set(open_tempfiles.keys())):
				print(list(set(open_chromosome_queue)))
				print(open_tempfiles.keys())
				raise Exception('There is a mismatch between the the open chromosome queue and the dictionary of open filehandles')

	for i in open_tempfiles:  # closing all the output files
		open_tempfiles[i].close()
	return chromosomes		  # this returns all the chromosome names and temp output file names so that one can keep track of how many tmp files there are and what the file names are.

def splitString(string,step=50):
	""" splits a string in blocks of step. Useful for chopping up sequences into smaller pieces """
	return "\n".join([string[i:i+step] for i in range(0,len(string),step)])

#------------------------------------------------------------------------------------------------------------
# SAM format functions
# No longer used because of pysam but still useful to have around
#------------------------------------------------------------------------------------------------------------

def flag2Bitstring(flag):
	""" decodes the SAM integer flag in the second column and returns a bit string """
	bitstring = str()
	bitvalues = [1024,512,256,128,64,32,16,8,4,2,1]		# The bitstrings are 11 characters long
	count_ones = 0
	for numbers in bitvalues:
		if flag - numbers < 0:
			bitstring += '0'
			continue
		if flag - numbers > 0:
			bitstring += '1'
			count_ones += 1
			flag = flag - numbers
			continue
		if flag - numbers == 0:
			bitstring += '1'
			count_ones += 1
			flag = flag - numbers
			continue
		if flag == 0:
			bitstring += '0'

	if count_ones > 0:
		return bitstring
	else:
		return None

def decodeBitstring(bitstring):
	""" decodes bitstrings generated by the flag2bitstring method """
	dictionary = {
		'PAIRED'	   : False,
		'PROPER_PAIR'  : False,
		'MAPPED_QUERY' : True,
		'MAPPED_MATE'  : True,
		'STRAND_QUERY' : "F",
		'STRAND_MATE'  : "R",
		'READ_TYPE'	   : "S",
		'HIT_REPEAT'   : "U",
		'READ_QUAL_CHECK' : True,
		'PCR_DUPLICATE': False
	}

	if bitstring is None:
		return None
	if bitstring[-1] == '1':	# The read is paired in sequencing no matter whether it is mapped in a paired
		dictionary['PAIRED'] = True
		# then if the read is indeed paired the following needs to be tested:
		if bitstring[-2] == '0':	# The read is mapped in a proper pair (depends on the protocol, normally inferred during alignment)
			dictionary['PROPER_PAIR'] = True
		if bitstring[-3] == '1':	# The query sequence itself is unmapped
			dictionary['MAPPED_QUERY'] = False
		if bitstring[-4] == '1':	# The mate sequence is unmapped
			dictionary['MAPPED_MATE'] = False
		if bitstring[-6] == '0':	# The strand of the mate
			dictionary['STRAND_MATE'] = "F"
		if bitstring[-7] == '1':	# The read is the first read in the paired
			dictionary['READ_TYPE'] = "L"
		if bitstring[-8] == '1':	# The read is the second read in the paired
			dictionary['READ_TYPE'] = "R"
	if bitstring[-5] == '1':	# The strand of the query
		dictionary['STRAND_QUERY'] = "R"
	if bitstring[-9] == '1':	# The read has multiple alignment locations in the genome
		dictionary['HIT_REPEAT'] = "R"
	if bitstring[-10] == '1':	# The read read fails vendor quality check
		dictionary['READ_QUAL_CHECK'] = False
	if bitstring[-11] == '1':	# The read read is a PCR or an optical duplicate
		dictionary['PCR_DUPLICATE'] = True

	return dictionary

#------------------------------------------------------------------------------------------------------------
# DNA and RNA manipulation functions
#------------------------------------------------------------------------------------------------------------

def complement(sequence):
	""" returns the complement of the DNA sequence """
	basecomplement = str()
	complement = {"A":"T","G":"C","C":"G","T":"A","a":"t","c":"g","g":"c","t":"a"}
	letters	 = list(sequence)
	for letter in letters:
		if letter in complement:
			basecomplement += complement[letter]
		else:
			basecomplement += letter
	return basecomplement

def reverse_complement(sequence):
	""" Returns the reverse complement of a DNA string """
	basecomplement = str()
	complement = {"A":"T","G":"C","C":"G","T":"A","a":"t","c":"g","g":"c","t":"a"}
	letters	 = list(sequence)
	for letter in letters:
		if letter in complement:
			basecomplement += complement[letter]
		else:
			basecomplement += letter
	return basecomplement[::-1]

def translate(sequence,type="DNA",frame=1):
	""" Translates a DNA or RNA sequence into a protein sequence. Type variable specifies whether the source is DNA (type='DNA') or RNA (type='RNA') """
	assert type in ["DNA","RNA"], "supported types are 'DNA' or 'RNA'\n"
	protein	 = str()
	codonmap = {"UUU":"F", "UUC":"F", "UUA":"L", "UUG":"L",
				"UCU":"S", "UCC":"S", "UCA":"S", "UCG":"S",
				"UAU":"Y", "UAC":"Y", "UAA":"*", "UAG":"*",
				"UGU":"C", "UGC":"C", "UGA":"*", "UGG":"W",
				"CUU":"L", "CUC":"L", "CUA":"L", "CUG":"L",
				"CCU":"P", "CCC":"P", "CCA":"P", "CCG":"P",
				"CAU":"H", "CAC":"H", "CAA":"Q", "CAG":"Q",
				"CGU":"R", "CGC":"R", "CGA":"R", "CGG":"R",
				"AUU":"I", "AUC":"I", "AUA":"I", "AUG":"M",
				"ACU":"T", "ACC":"T", "ACA":"T", "ACG":"T",
				"AAU":"N", "AAC":"N", "AAA":"K", "AAG":"K",
				"AGU":"S", "AGC":"S", "AGA":"R", "AGG":"R",
				"GUU":"V", "GUC":"V", "GUA":"V", "GUG":"V",
				"GCU":"A", "GCC":"A", "GCA":"A", "GCG":"A",
				"GAU":"D", "GAC":"D", "GAA":"E", "GAG":"E",
				"GGU":"G", "GGC":"G", "GGA":"G", "GGG":"G",}

	sequence = list(sequence.upper())
	if type == "DNA":
		for i,nuc in enumerate(sequence):
			if nuc == "T":
				sequence[i] = "U"
		sequence = "".join(sequence)
	for i in range(frame-1,len(sequence),3):        # find codons in the sequence starting at position 1 in sequence
		codon = sequence[i:i+3]
		if not protein and codonmap[codon] == "M":  # if the start codon has been found
			protein += codonmap[codon]
		elif protein and len(codon) != 3:
			sys.stderr.write("\nWarning! Protein is a truncated sequence\n")
			break
		elif protein and codonmap[codon]:
			protein += codonmap[codon]
		elif protein and codonmap[codon] == "STOP":
			break
	return protein

#------------------------------------------------------------------------------------------------------------

def basequality(letter,ftype="L"):
	""" Returns a phred or solexa base quality from a character. The ftype indicates 'sanger' or 'solexa' or 'illumina'
	S - Sanger		  Phred+33,	 raw reads typically (0, 40)
	X - Solexa		  Solexa+64, raw reads typically (-5, 40)
	I - Illumina 1.3+ Phred+64,	 raw reads typically (0, 40)
	J - Illumina 1.5+ Phred+64,	 raw reads typically (3, 40) with 0=unused, 1=unused, 2=Read Segment Quality Control Indicator (bold)
	L - Illumina 1.8+ Phred+33,	 raw reads typically (0, 41)
	Default = L"""
	setone = ["L","S"]
	settwo = ["X","J","I"]
	if ftype in setone:
		return ord(letter) - 33
	elif ftype in settwo:
		return ord(letter) - 64
	else:
		raise RuntimeError("could not correctly interpret the file type of the raw dataset\nPlease use phred,solexa or illumina\n")

def getfilename(filepath):
	""" Returns the file name without path and extension"""
	if isinstance(filepath,file_types):
		return "standard_input"
	else:
		return os.path.splitext(os.path.basename(filepath))[0]

def splitter(line):
	""" This function splits the line at double quotes and returns the second object in the array"""
	line = line.replace('\"',"")
	return line.split()[1]

def check_list(list_input,key): # Initially I thought of using the builtin python 'set' module and its add() method but the check_list runs significantly faster when used with lists!
	""" This function checks whether a key already exists in a list.  is often used in this module to make sure that lists do not contain duplicate gene or ORF"""
	if key not in list_input:
		list_input.append(key)
		return True
	else:
		return False

def find_key(dic, val):
	""" Return the key of dictionary dic given the value"""
	return [k for k, v in six.iteritems(dic) if v == val][0]

def readisDuplicate(chromosome,tuple_input,tracker_dict,activate=False):
	""" Takes a set() list and checks whether tuples with read_coordinates are already present
	in the set. If not, it will add the coordinates to the list, else it will return False.
	The reason for using sets is because membership tests are much faster than regular python lists and
	the list doesn't need to be ordered."""
	input = str(tuple_input)  # making a string out of the input coordinates makes comparison a lot easier/faster
	if activate is False:
		return False
	elif input in tracker_dict[chromosome]:
		return True
	else:
		tracker_dict[chromosome].add(input)
		return False

#------------------------------------------------------------------------------------------------------------
# STRAND MANIPULATION FUNCTIONS
#------------------------------------------------------------------------------------------------------------

def orientation(read_strand,gene_strand):
	""" returns sense if read and gene strand are the same. If not it returns anti-sense.
	if the interval has no strand (i.e. '.') then it returns sense. """
	if read_strand == "F": read_strand = "+"
	if read_strand == "R": read_strand = "-"
	if read_strand == "." or gene_strand == ".":
		return "sense"
	elif read_strand == gene_strand:
		return "sense"
	else:
		return "anti-sense"

def strand_converter(strand):
	""" Converts the novoalign strand characters to GTF strand characters (i.e. 'F' to '+' and 'R' to '-')"""
	if not strand:
		return None
	elif strand == "F":
		return "+"
	elif strand == "R":
		return "-"
	# just in case the strand is already + or -, then return the strand itself:
	elif strand == "-" or strand == "+":
		return strand
	else:
		raise TypeError("\nThe strand information could not be interpreted.\nPlease verify your input\n")

def reverse_strand(strand):
	""" Reverses the strand"""
	if	 strand == "F": return "R"
	elif strand == "R": return "F"
	elif strand == "+": return "-"
	elif strand == "-": return "+"
	else:
		raise TypeError("\nThe strand information could not be interpreted.\nPlease verify your input\n")

def get_sequence(chromosome,strand,upstream,downstream,tab_dict):
	""" Returns the nucleotide sequence from "chromosome" using upstream and downstream coordinates.Usage: get_sequence(chromosome,strand,upstream,downstream,dictionary)"""
	if strand == "+" or strand == "F" and tab_dict[chromosome][upstream:downstream]:
		return tab_dict[chromosome][upstream:downstream]
	elif strand == "-" or strand == "R" and tab_dict[chromosome][upstream:downstream]:
		return reverse_complement(tab_dict[chromosome][upstream:downstream])
	else:
		raise NoSequenceError("\ncould not find a sequence for chromosome %s and coordinates %d-%d, please check your input\n" %(chromosome,upstream,downstream))

#------------------------------------------------------------------------------------------------------------
# OVERLAP FUNCTIONS
#------------------------------------------------------------------------------------------------------------

def calc_list_overlap(coordinates,gene_iter_coordinates,list_type="read"):
	""" Returns a numpy array containing zeros and ones, ones indicate the position in the gene where reads were found to overlap with
	locations of mutations. List_type 'read' indicates that the coordinates array contains read coordinates, wherease list_type 'muts'
	indicates an array with genomic locations of mutations, such as deletions or substitutions"""
	if isinstance(gene_iter_coordinates,np.ndarray):
		if list_type == "read":
			return np.where((gene_iter_coordinates >= coordinates[0]) & (gene_iter_coordinates <= coordinates[1]),1,0)
		elif list_type == "muts":
			array = np.zeros(len(gene_iter_coordinates),dtype=int)
			for position in coordinates:
				array += np.where((gene_iter_coordinates == position),1,0)
			return array
		else:
			raise TypeError("please specify whether the first array contains read coordinates (i.e. list_type='read') or a list of mutations (i.e. list_type='muts')\n")
	else:
		raise TypeError("your gene range coordinates are not numpy arrays! Please double check your input\n")

def intron_overlap(read_start,read_end,intron_coordinates,sequence):
	""" Determines whether a read overlaps introns"""
	overlap = False
	if sequence == "genomic":
		overlap = False
	if sequence == "coding" and intron_coordinates:
		copy_b = list()
		copy_b.extend(intron_coordinates)
		while copy_b:
			intron_coord = copy_b.pop()
			intron_down = intron_coord[1]
			intron_up = intron_coord[0]
			if read_end <= intron_down and read_start >= intron_up:
				overlap = True
			if read_end >= intron_down and read_start <= intron_down:
				overlap = True
			if read_start <= intron_up and read_end >= intron_up:
				overlap = True
	return overlap

def old_overlap(a, b, c, d):	# first overlap function. Slowest of all three
	""" Checks whether there is overlap between read coordinates and chromosomal coordinates.
	a and b are read coordinates, whereas c and d are gene coordinates"""
	if a >= c and  b <= d:
		return True
	elif a < c and b <= d and b >= c:
		return True
	elif a >= c and b > d and a <= d:
		return True
	elif a < c and b >= d:
		return True
	else:
		return False

def overlap(a, b, c, d, overlap=1):		# improved overlap function. Marginally faster than the first one
	""" Checks whether there is overlap between read coordinates and chromosomal coordinates.
	a and b are read coordinates, whereas c and d are gene coordinates. The read or cluster can
	be larger than the gene itself. The result of this function can return zero, which means that two positions are identical.
	This is why the overlap value should always be subtracted by 1.

	Possibilities:

	   start	   start				   end						   start				  end	   start				   end
	 ---|			|----------------------|							|----------------------|		|----------------------|
		|---					<-overlap->								<-overlap->						<-------overlap-------->
	   end	(overlap=0)	   f0					f1		 f0						   f1				 f0							   f1
							|-------------------|		  |------------------------|				  |----------------------------|

	f0					  f1							   f0					   f1	   f0					  f1
	 |----------------------|								|----------------------|		|----------------------|
				  <-overlap->								 <-overlap->					 <-------overlap-------->
				 start				 end	 start					  end				 start						  end
				   |------------------|		  |------------------------|				  |----------------------------|

	Won't work for the following:

	 f0			   f1			 start			end			start		   end			  f0			f1
	 |--------------| <-overlap-> |--------------|			  |-------------| <-overlap-> |-------------|

	"""
	overlap -= 1
	if b - c >= overlap and d - a >= overlap:
		return True
	else:
		return False

def numpy_overlap(numpy_array,start,end,overlap=1,returnall=False):	# much improved overlap function which uses numpy numerical calculations. Three to four times faster than the old ones
	""" Checks whether there is overlap between read coordinates and chromosomal coordinates but uses numpy arrays
	and numpy numerical conversions. Is much faster than regular iterations over lists. The read or cluster can be larger than the
	gene itself. The result of the overlap function can return zero, which means that two positions in the intervals are identical.
	This is why the overlap value should always be subtracted by 1.

	Possibilities:

	   start	   start				   end						   start				  end	   start				   end
	 ---|			|----------------------|							|----------------------|		|----------------------|
		|---					<-overlap->								<-overlap->						<-------overlap-------->
	   end	(overlap=0)	   f0					f1		 f0						   f1				 f0							   f1
							|-------------------|		  |------------------------|				  |----------------------------|

	f0					  f1							   f0					   f1	   f0					  f1
	 |----------------------|								|----------------------|		|----------------------|
				  <-overlap->								 <-overlap->					 <-------overlap-------->
				 start				 end	 start					  end				 start						  end
				   |------------------|		  |------------------------|				  |----------------------------|

	Won't work for the following:

	 f0			   f1			 start			end			start		   end			  f0			f1
	 |--------------| <-overlap-> |--------------|			  |-------------| <-overlap-> |-------------|

	"""
	overlap -= 1
	seqoverlap = (end - numpy_array['f0'] >= overlap) & (numpy_array['f1'] - start >= overlap)	# determines whether there is overlap between a read and gene chromosomal coordinates. Returns a bool index.

	if returnall:
		return list(numpy_array[seqoverlap])
	else:
		return list(numpy_array[seqoverlap]['f2'])		# if overlap was found, then the gene names are stored in the results list, otherwise it returns an empty list

def read_seq_overlap(seq_a,start_a,seq_b,start_b):
	end_a = start_a + len(seq_a)
	end_b = start_b + len(seq_b)
	if start_a >= start_b and start_a <= end_b:
		return ((start_a - start_b), True)
	elif start_a <= start_b and end_a >= start_b:
		return ((start_b - start_a), True)
	elif start_a < start_b and end_a < end_b:
		return ((start_b - end_a), False)
	elif start_a > start_b and end_a > end_b:
		return ((start_a - end_b), False)
	elif start_a == start_b:
		return (0,True)
	else:
		return (None, False)

#------------------------------------------------------------------------------------------------------------

def sortbyvalue(d):
	""" This function takes a dictionary and generates a list in which the keys are sorted by their value"""
	items = [(v, k) for k, v in list(d.items())]
	items.sort()
	items.reverse()				# so largest is first
	items = [(k, v) for v, k in items]
	return items

def sortbylistlength(d):
	""" This function takes a dictionary and generates a list in which the keys are sorted by their value"""
	items = [(len(v), k) for k, v in list(d.items())]
	items.sort()
	items.reverse()				# so largest is first
	items = [(k, v) for v, k in items]
	return items

def is_numeric(obj):
	""" Checks whether a list or numpy array has numbers """
	attrs = ['__add__', '__sub__', '__mul__', '__div__', '__pow__']
	return all(hasattr(obj, attr) for attr in attrs)
